from naeural_core.constants import COMMANDS, CONFIG_STREAM, PAYLOAD_DATA

class _CmdAPIMixin(object):

  def __init__(self):
    self._commands = []
    super(_CmdAPIMixin, self).__init__()
    return


  def _cmdapi_refresh(self):
    self._commands = []
    return
  

  def get_commands_after_exec(self):
    c = self._commands
    self._cmdapi_refresh()
    return c  
  
  
  def _cmdapi_send_commands(self):
    commands = self.get_commands_after_exec()
    if len(commands) > 0:
      self.commands_deque.append(commands)    
    return  
  

  def cmdapi_register_command(
    self, 
    node_address, 
    command_type, 
    command_content,
    send_immediately=True,
  ):
    """
    Send a command to a particular Execution Engine

    Parameters
    ----------
    node_address : str
      target Execution Engine.
      
    command_type : st
      type of the command - can be one of 'RESTART','STATUS', 'STOP', 'UPDATE_CONFIG', 
      'DELETE_CONFIG', 'ARCHIVE_CONFIG', 'DELETE_CONFIG_ALL', 'ARCHIVE_CONFIG_ALL', 'ACTIVE_PLUGINS', 
      'RELOAD_CONFIG_FROM_DISK', 'FULL_HEARTBEAT', 'TIMERS_ONLY_HEARTBEAT', 'SIMPLE_HEARTBEAT',
      'UPDATE_PIPELINE_INSTANCE', etc.
      
    command_content : dict
      the actual content - can be None for some commands.

    Returns
    -------
    None.

    """
    if isinstance(command_content, dict) and PAYLOAD_DATA.TIME not in command_content:
      command_content[PAYLOAD_DATA.TIME] = self.log.now_str(nice_print=True)
    box_id = self.net_mon.network_node_eeid(node_address)
    self._commands.append((box_id, node_address, self.use_local_comms_only, command_type, command_content))
    if send_immediately:
      self.P("CMDAPI: Sending async command '{}' for node '{}' <{}>".format(command_type, box_id, node_address))
      self._cmdapi_send_commands()
    else:
      self.P("CMDAPI: Register (after process send) command '{}' for node '{}' <{}>".format(command_type, box_id, node_address))
    return
  
  ###
  ### START Official API 
  ### 
  
  def cmdapi_start_pipeline(self, config, node_address=None):
    """
    Sends a start pipeline to a particular destination Execution Engine

    Parameters
    ----------
    node_address : str, optional
      destination Execution Engine, `None` will default to local Execution Engine. The default is None
      .
    config : dict
      the pipeline configuration. 

    Returns
    -------
    None.
    
    Example:
      
      ```
      config = {
        "NAME" : "test123",
        "TYPE" : "Void",
      }
      ee_id = None # using current processing node
      plugin.cmdapi_start_pipeline(config=config, node_address=ee_addr)
      ```

    """
    self._cmdapi_start_stream_by_config(config_stream=config, node_address=node_address)
    return
  
  
  def cmdapi_update_instance_config(
    self, pipeline, signature, instance_id, instance_config, node_address=None,
    send_immediately=True,
  ):
    """
    Sends update config for a particular plugin instance in a given box/pipeline
    

    Parameters
    ----------
      
    pipeline : str
      Name of the pipeline. 
      
    signature: str
      Name (signature) of the plugin 
      
    instance_id: str
      Name of the instance
      
    instance_config: dict
      The configuration for the given box/pipeline/plugin/instance

    node_address : str, optional
      destination Execution Engine, `None` will default to local Execution Engine. The default is None.

    Returns
    -------
    None.
    
    """
    self._cmdapi_update_pipeline_instance(
      pipeline=pipeline,
      signature=signature,
      instance_id=instance_id,
      instance_config=instance_config,
      node_address=node_address,
      send_immediately=send_immediately,
    )
    return
  
  
  def cmdapi_batch_update_instance_config(
    self, lst_updates, node_address=None,
    send_immediately=True,
  ):
    """Send a batch of updates for multiple plugin instances within their individual pipelines

    Parameters
    ----------
    lst_updates : list of dicts
        The list of updates for multiple plugin instances within their individual pipelines
        
    node_address : str, optional
        Destination node, by default None
        
    Returns
    -------
    None.
    
    Example
    -------
    
      ```python
      # in this example we are modifying the config for 2 instances of the same plugin `A_PLUGIN_01`
      # within the same pipeline `test123`
      lst_updates = [
        {
          "NAME" : "test123",
          "SIGNATURE" : "A_PLUGIN_01",
          "INSTANCE_ID" : "A_PLUGIN_01_INSTANCE_01",
          "INSTANCE_CONFIG" : {
            "PARAM1" : "value1",
            "PARAM2" : "value2",
          }
        },
        {
          "NAME" : "test123",
          "SIGNATURE" : "A_PLUGIN_01",
          "INSTANCE_ID" : "A_PLUGIN_01_INSTANCE_02",
          "INSTANCE_CONFIG" : {
            "PARAM1" : "value1",
            "PARAM2" : "value2",
          }
        },
      ] 
      plugin.cmdapi_batch_update_instance_config(lst_updates=lst_updates, node_address=None)
      ```
    
    """
    # first check if all updates are valid
    for update in lst_updates:
      assert isinstance(update, dict), "All updates must be dicts"
      assert PAYLOAD_DATA.NAME in update, "All updates must have a pipeline name"
      assert PAYLOAD_DATA.SIGNATURE in update, "All updates must have a plugin signature"
      assert PAYLOAD_DATA.INSTANCE_ID in update, "All updates must have a plugin instance id"
      assert PAYLOAD_DATA.INSTANCE_CONFIG in update, "All updates must have a plugin instance config"
      assert isinstance(update[PAYLOAD_DATA.INSTANCE_CONFIG], dict), "All updates must have a plugin instance config as dict"
    #endfor check all updates
    self.cmdapi_register_command(
      node_address=node_address,
      command_type=COMMANDS.BATCH_UPDATE_PIPELINE_INSTANCE,
      command_content=lst_updates,
      send_immediately=send_immediately,
    )
    return
  
  
  def cmdapi_send_instance_command(self, pipeline, signature, instance_id, instance_command, node_address=None):
    """
    Sends a INSTANCE_COMMAND for a particular plugin instance in a given box/pipeline

    Parameters
    ----------
    pipeline : str
      Name of the pipeline. 
      
    instance_id: str
      Name of the instance
      
    signature: str
      Name (signature) of the plugin     
      
    instance_command: any
      The configuration for the given box/pipeline/plugin/instance. Can be a string, dict, etc

    node_address : str, optional
      destination Execution Engine, `None` will default to local Execution Engine. The default is None.
      
      
    Returns
    -------
    None.
    
    Example:
    --------
    
    
      ```
      pipeline = "test123"
      signature = "A_PLUGIN_01"
      instance_id = "A_PLUGIN_01_INSTANCE_01"
      instance_command = {
        "PARAM1" : "value1",
        "PARAM2" : "value2",
      }
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=None,
      )
      ```
        
    """
    instance_config = {
      'INSTANCE_COMMAND' : instance_command
    }
    self._cmdapi_update_pipeline_instance(
      pipeline=pipeline,
      signature=signature,
      instance_id=instance_id,
      instance_config=instance_config,
      node_address=node_address,
    )
    return
  
  
  def cmdapi_archive_pipeline(self, node_address=None, name=None):
    """
    Stop and archive a active pipeline on destination Execution Engine

    Parameters
    ----------
    node_address : str, optional
      destination Execution Engine, `None` will default to local Execution Engine. The default is None.
    name : str, optional
      Name of the pipeline. The default is `None` and will point to current pipeline where the plugin instance 
      is executed.

    Returns
    -------
    None.

    """
    self._cmdapi_archive_stream(node_address=node_address, stream_name=name)
    return


  def cmdapi_stop_pipeline(self, node_address=None, name=None):
    """
    Stop and delete a active pipeline on destination Execution Engine

    Parameters
    ----------
    node_address : str, optional
      destination Execution Engine, `None` will default to local Execution Engine. The default is None.
    name : str, optional
      Name of the pipeline. The default is `None` and will point to current pipeline where the plugin instance 
      is executed.

    Returns
    -------
    None.

    """
    self._cmdapi_stop_stream(node_address=node_address, stream_name=name)
    return


  def cmdapi_archive_all_pipelines(self, node_address=None):
    """
    Stop all active pipelines on destination Execution Engine

    Parameters
    ----------
    node_address : str, optional
      Address of the target E2 instance. The default is `None` and will run on local E2.

    Returns
    -------
    None.

    """
    node_address = node_address or self.node_addr
    self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.ARCHIVE_CONFIG_ALL, command_content=None)
    return
  
  
  def cmdapi_start_pipeline_by_params(self, name, pipeline_type, node_address=None, url=None,
                                      reconnectable=None, live_feed=False, plugins=None,
                                      stream_config_metadata=None, cap_resolution=None, 
                                      **kwargs):
    """
    Start a pipeline by defining specific pipeline params

    Parameters
    ----------
    name : str
      Name of the pipeline.
      
    pipeline_type : str
      type of the pipeline. Will point the E2 instance to a particular Data Capture Thread plugin
      
    node_address : str, optional
      Address of the target E2 instance. The default is `None` and will run on local E2.
      
    url : str, optional
      The optional URL that can be used by the DCT to acquire data. The default is None.
      
    reconnectable : str, optional
      Attempts to reconnect after data stops if 'YES'. 'KEEP_ALIVE' will not reconnect to
      data source but will leave the DCT in a "zombie" state waiting for external pipeline 
      close command.
      The default is 'YES'.
      
    live_feed : bool, optional
      Will always try to generate the real-time datapoint (no queued data). The default is False.
      
    plugins : list of dicts, optional
      Lists all the business plugins with their respective individual instances. The default is None.
      
    stream_config_metadata : dict, optional
      Options (custom) for current DCT. The default is None.
      
    cap_resolution : float, optional
      Desired frequency (in Hz) of the DCT data reading cycles. The default is None.
      

    Returns
    -------
    None (actually)
    
    Example
    -------
    
      ```
      name = "test123"
      pipeline_type = "Void"
      plugins = [
        {
          "SIGNATURE" : "A_PLUGIN_01",
          "INSTANCES" : [
            {
              "INSTANCE_ID" : "A_PLUGIN_01_INSTANCE_01",
              "PARAM1" : "value1",
              "PARAM2" : "value2",
            }
          ]
        }
      ]
      plugin.cmdapi_start_pipeline_by_params(
        name=name, 
        pipeline_type=pipeline_type, 
        plugins=plugins,
      )
      ```

    """
    return self._cmdapi_start_stream_by_params(
      name=name, stream_type=pipeline_type, node_address=node_address, url=url, reconnectable=reconnectable,
      live_feed=live_feed, plugins=plugins, stream_config_metadata=stream_config_metadata,
      cap_resolution=cap_resolution, **kwargs
    )
  
  
  def cmdapi_start_simple_custom_pipeline(self, *, base64code, node_address=None, name=None, instance_config={}, **kwargs):
    """
    Starts a CUSTOM_EXEC_01 plugin on a Void pipeline
    

    Parameters
    ----------
    base64code : str
      The base64 encoded string that will be used as custom exec plugin.
      
    node_address : str, optional
      The destination processing node. The default is None and will point to current node.
      
    name : str, optional
      Name of the pipeline. The default is None and will be uuid generated.
      
    instance_config / kwargs: dict
      Dict with params for the instance that can be given either as a dict or as kwargs


    Returns
    -------
    name : str
      returns the name of the pipeline.
      
    
    Example
    -------
    
      ```
      worker = plugin.cfg_destination                     # destination worker received in plugin json command
      worker_code = plugin.cfg_worker_code                # base64 code that will be executed 
      custom_code_param = plugin.cfg_custom_code_param    # a special param expected by the custom code
      pipeline_name = plugin.cmdapi_start_simple_custom_pipeline(
        base64code=worker_code, 
        node_address=worker,
        custom_code_param=pcustom_code_param,
      )
      ```
    """
    if name is None:
      name = self.uuid(8)
    plugins = [
      {
        self.const.BIZ_PLUGIN_DATA.INSTANCES: [
          {
            self.const.BIZ_PLUGIN_DATA.INSTANCE_ID: "SIMPLE_C_EXEC",
            "CODE": base64code,
            **instance_config,
            **kwargs,
          }
        ],
        self.const.BIZ_PLUGIN_DATA.SIGNATURE: "CUSTOM_EXEC_01",
      }
    ]
    self.cmdapi_start_pipeline_by_params(
        name=name, 
        pipeline_type="Void", 
        node_address=node_address, 
        plugins=plugins,
    )
    return name
  
  
  def cmdapi_send_pipeline_command(self, command, node_address=None, pipeline_name=None):
    """Sends a command to a particular pipeline on a particular destination E2 instance

    Parameters
    ----------
    command : any
        the command content

    node_address : str, optional
        name of the destination e2, by default None (self)

    pipeline_name : str, optional
        name of the pipeline, by default None (self)
        
    
    Returns
    -------
    None.
    
    Example
    -------
    
      ```
      # send a command directly to the current pipeline
      plugin.cmdapi_send_pipeline_command(
        command={"PARAM1" : "value1", "PARAM2" : "value2"},
        node_address=None,
        pipeline_name=None,
      )
      ```
    
    """
    node_address = node_address or self.node_addr
    pipeline_name = pipeline_name or self.get_stream_id()
    payload = {
      PAYLOAD_DATA.NAME : pipeline_name,
      COMMANDS.PIPELINE_COMMAND : command,
    }
    self.cmdapi_register_command(
      node_address=node_address,
      command_type=COMMANDS.PIPELINE_COMMAND,
      command_content=payload
    )    
    return
  ###
  ### END Official API 
  ### 




  # STOP BOX SECTION
  if True:
    def _cmdapi_stop_box(self, node_address=None):
      node_address = node_address or self.node_addr
      self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.STOP, command_content=None)
      return

    def cmdapi_stop_current_box(self):
      self._cmdapi_stop_box(node_address=None)
      return

    def cmdapi_stop_other_box(self, node_address):
      self._cmdapi_stop_box(node_address=node_address)
      return
  #endif

  # RESTART BOX SECTION
  if True:
    def _cmdapi_restart_box(self, node_address=None):
      node_address = node_address or self.node_addr
      self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.RESTART, command_content=None)
      return

    def cmdapi_restart_current_box(self):
      self._cmdapi_restart_box(node_address=None)
      return

    def cmdapi_restart_other_box(self, node_address):
      self._cmdapi_restart_box(node_address=node_address)
      return
  #endif

  # START STREAM SECTION
  if True:
    def _cmdapi_start_stream_by_config(
      self, config_stream, node_address=None,
      send_immediately=True,
    ):
      node_address = node_address or self.node_addr

      for param in CONFIG_STREAM.MANDATORY:
        if param not in config_stream:
          self.P("CMDAPI: Param '{}' not configured for commanded stream. Cannot register command".format(
            param), color='r'
          )
          return

      self.cmdapi_register_command(
        node_address=node_address, command_type=COMMANDS.UPDATE_CONFIG, command_content=config_stream,
        send_immediately=send_immediately,
      )
      return

    def cmdapi_start_stream_by_config_on_current_box(self, config_stream):
      self._cmdapi_start_stream_by_config(config_stream=config_stream, node_address=None)
      return

    def cmdapi_start_stream_by_config_on_other_box(self, node_address, config_stream):
      self._cmdapi_start_stream_by_config(config_stream=config_stream, node_address=node_address)
      return

    # Metastreams
    def _cmdapi_start_metastream_by_config(self, config_metastream, node_address=None, collected_streams=None):
      config_metastream[CONFIG_STREAM.TYPE] = CONFIG_STREAM.METASTREAM
      if collected_streams is not None and isinstance(collected_streams, list) and len(collected_streams) > 0:
        config_metastream[CONFIG_STREAM.COLLECTED_STREAMS] = collected_streams
        CONFIG_STREAM.COLL
      self._cmdapi_start_stream_by_config(config_stream=config_metastream, node_address=node_address)
      return

    def cmdapi_start_metastream_by_config_on_current_box(self, config_metastream, collected_streams=None):
      self._cmdapi_start_metastream_by_config(config_metastream=config_metastream, node_address=None, collected_streams=collected_streams)
      return

    def cmdapi_start_metastream_by_config_on_other_box(self, node_address, config_metastream, collected_streams=None):
      self._cmdapi_start_metastream_by_config(config_metastream=config_metastream, node_address=node_address, collected_streams=collected_streams)
      return
    # end Metastreams


    def _cmdapi_start_stream_by_params(
      self, name, stream_type, url=None,
      reconnectable=None, live_feed=False, plugins=None,
      stream_config_metadata=None, cap_resolution=None, node_address=None, 
      send_immediately=True, 
      **kwargs
    ):

      config_stream = {
        CONFIG_STREAM.K_NAME          : name,
        CONFIG_STREAM.K_TYPE          : stream_type,
      }

      if reconnectable is not None:
        config_stream[CONFIG_STREAM.K_RECONNECTABLE] = reconnectable

      if live_feed is not None:
        config_stream[CONFIG_STREAM.K_LIVE_FEED] = live_feed

      if url is not None:
        config_stream[CONFIG_STREAM.K_URL] = url

      if plugins is not None:
        config_stream[CONFIG_STREAM.K_PLUGINS] = plugins

      if stream_config_metadata is not None:
        config_stream[CONFIG_STREAM.STREAM_CONFIG_METADATA] = stream_config_metadata

      if cap_resolution is not None:
        config_stream[CONFIG_STREAM.CAP_RESOLUTION] = cap_resolution
      
      config_stream = {
        **config_stream, 
        **{k.upper():v for k,v in kwargs.items()},
      }
      self._cmdapi_start_stream_by_config(
        config_stream=config_stream, node_address=node_address,
        send_immediately=send_immediately,
      )
      return


    def cmdapi_start_stream_by_params_on_current_box(self, name, stream_type, url=None,
                                                     reconnectable=None, live_feed=False, plugins=None,
                                                     stream_config_metadata=None, cap_resolution=None, **kwargs):
      self._cmdapi_start_stream_by_params(
        name=name, stream_type=stream_type, url=url,
        reconnectable=reconnectable, live_feed=live_feed, plugins=plugins,
        stream_config_metadata=stream_config_metadata, cap_resolution=cap_resolution,
        node_address=None, **kwargs
      )
      return

    def cmdapi_start_stream_by_params_on_other_box(
      self, node_address, name, stream_type, url=None,
      reconnectable=None, live_feed=False, plugins=None,
      stream_config_metadata=None, cap_resolution=None, 
      send_immediately=True, **kwargs
    ):
      self._cmdapi_start_stream_by_params(
        name=name, stream_type=stream_type, url=url,
        reconnectable=reconnectable, live_feed=live_feed, plugins=plugins,
        stream_config_metadata=stream_config_metadata, cap_resolution=cap_resolution,
        node_address=node_address, 
        send_immediately=send_immediately,
        **kwargs
      )
      return
    
    def _cmdapi_update_pipeline_instance(
      self, pipeline, signature, instance_id, instance_config, node_address=None,
      send_immediately=True,
    ):
      node_address = node_address or self.node_addr
      payload = {
        PAYLOAD_DATA.NAME : pipeline,
        PAYLOAD_DATA.SIGNATURE : signature,
        PAYLOAD_DATA.INSTANCE_ID : instance_id,
        PAYLOAD_DATA.INSTANCE_CONFIG : instance_config,
      }
      self.cmdapi_register_command(
        node_address=node_address,
        command_type=COMMANDS.UPDATE_PIPELINE_INSTANCE,
        command_content=payload,
        send_immediately=send_immediately,
      )
      return
  #endif

  # STOP STREAM SECTION
  if True:
    def _cmdapi_stop_stream(self, node_address=None, stream_name=None):
      node_address = node_address or self.node_addr
      stream_name = stream_name or self.get_stream_id()
      self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.DELETE_CONFIG, command_content=stream_name)
      return
    
    def cmdapi_stop_pipeline(self, node_address, name):
      self._cmdapi_stop_stream(node_address=node_address, stream_name=name)
      return
    
    def cmdapi_stop_current_pipeline(self):
      self._cmdapi_stop_stream(node_address=None, stream_name=None)
      return

    def cmdapi_stop_current_stream(self):
      self._cmdapi_stop_stream(node_address=None, stream_name=None)
      return

    def cmdapi_stop_other_stream_on_current_box(self, stream_name):
      self._cmdapi_stop_stream(node_address=None, stream_name=stream_name)
      return

    def cmdapi_stop_stream_on_other_box(self, node_address, stream_name):
      self._cmdapi_stop_stream(node_address=node_address, stream_name=stream_name)
      return
  #endif

  # ARCHIVE STREAM SECTION
  if True:
    def _cmdapi_archive_stream(self, node_address=None, stream_name=None):
      node_address = node_address or self.node_addr
      stream_name = stream_name or self.get_stream_id()
      self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.ARCHIVE_CONFIG, command_content=stream_name)
      return

    def cmdapi_archive_current_stream(self):
      self._cmdapi_archive_stream(node_address=None, stream_name=None)
      return

    def cmdapi_archive_other_stream_on_current_box(self, stream_name):
      self._cmdapi_archive_stream(node_address=None, stream_name=stream_name)
      return

    def cmdapi_archive_stream_on_other_box(self, node_address, stream_name):
      self._cmdapi_archive_stream(node_address=node_address, stream_name=stream_name)
      return
  #endif

  # FINISH STREAM ACQUISITION SECTION
  if True:
    def _cmdapi_finish_stream_acquisition(self, node_address=None, stream_name=None):
      node_address = node_address or self.node_addr
      stream_name = stream_name or self.get_stream_id()
      
      delta_config = {
        CONFIG_STREAM.NAME : stream_name,
        COMMANDS.COMMANDS : [{COMMANDS.FINISH_ACQUISITION : True}]
      }

      self.cmdapi_register_command(node_address=node_address, command_type=COMMANDS.UPDATE_CONFIG, command_content=delta_config)
      return

    def cmdapi_finish_current_stream_acquisition(self):
      self._cmdapi_finish_stream_acquisition(node_address=None, stream_name=None)
      return

    def cmdapi_finish_other_stream_acquisition_on_current_box(self, stream_name):
      self._cmdapi_finish_stream_acquisition(node_address=None, stream_name=stream_name)
      return

    def cmdapi_finish_stream_acquisition_on_other_box(self, node_address, stream_name):
      self._cmdapi_finish_stream_acquisition(node_address=node_address, stream_name=stream_name)
      return
  #endif




import traceback

from naeural_core import constants as ct

from naeural_core.data import CaptureManager
from naeural_core.comm import CommunicationManager
from naeural_core.config import ConfigManager
from naeural_core.serving import ServingManager
from naeural_core.business import BusinessManager
from naeural_core.io_formatters import IOFormatterManager
from naeural_core.heavy_ops import HeavyOpsManager
from naeural_core.remote_file_system import FileSystemManager
from naeural_core.bc import DefaultBlockEngine

from naeural_core.ipfs import R1FSEngine

class _ManagersInitMixin(object):
  def __init__(self):
    super(_ManagersInitMixin, self).__init__()
    return

  def _initialize_config_manager(self):
    header = "Initializing config manager ..."
    self.P(header, color='b', boxed=True)

    self._config_manager = ConfigManager(
      log=self.log,
      shmem=self.app_shmem,
      owner=self,
      subfolder_path=ct.CONFIG_MANAGER.DEFAULT_SUBFOLDER_PATH,
      fn_app_config=ct.CONFIG_MANAGER.DEFAULT_FN_APP_CONFIG,
      folder_streams_configs=ct.CONFIG_MANAGER.DEFAULT_FOLDER_STREAMS_CONFIGS
    )
    
    # maybe delete admin_pipeline
    
    if self.cfg_reset_admin_pipeline:
      self.P("Resetting admin pipeline...")
      self._config_manager._delete_stream_config('admin_pipeline')
    else:
      self.P("Preserving admin pipeline. Please make sure this is intended.", color='r', boxed=True)
    
    # first try to retrieve the config including the streams
    self.P("Initiating ConfigManager retrieve...")
    self._config_manager.retrieve(lst_config_retrieve=self.cfg_config_retrieve)
    try:
      # then load the config_app.txt or similar and then the streams
      self.P("Initiating ConfigManager load...")      
      self._config_manager.load()
    except Exception as e:
      self.P(str(e))
      raise e
    return

  def _initialize_comm_manager(self):
    avail_cmds = self.get_commands_list()
    header = "Initializing comm mgr ..."
    self.P(header, color='b', boxed=True)
    self.P(" Comm mgr commands: {}".format(avail_cmds))

    self._comm_manager = CommunicationManager(
      log=self.log,
      config=self._config_manager.cfg_app_communication,
      shmem=self.app_shmem,
      device_version=self.__version__,
      environment_variables=self.cfg_communication_environment,
      avail_commands=avail_cmds,
      DEBUG=self.DEBUG
    )
    self._comm_manager.validate(raise_if_error=True)
    self._comm_manager.start_communication()
    
    self.app_shmem["comm_manager"] = self._comm_manager
    return

  def _initialize_capture_manager(self):
    header = "Initializing capture manager ..."
    self.P(header, color='b', boxed=True)
    
    self._capture_manager = CaptureManager(
      log=self.log, 
      shmem=self.app_shmem, 
      owner=self,
      DEBUG=self.DEBUG,
      environment_variables=self.cfg_capture_environment,
    )

    self.app_shmem["capture_manager"] = self._capture_manager
    return

  def _initialize_serving_manager(self):
    header = "Initializing serving manager ..."
    self.P(header, color='b', boxed=True)

    self._serving_manager = ServingManager(
      log=self.log,
      shmem=self.app_shmem,
      DEBUG=self.DEBUG,
      full_debug=self.cfg_serving_environment.get('FULL_DEBUG', False),
      log_timeouts_period=self.cfg_serving_environment.get('LOG_TIMEOUTS_PERIOD', 3600),
      owner=self,
    )

    self.app_shmem["serving_manager"] = self._serving_manager
    return

  def _initialize_business_manager(self):
    header = "Initializing business manager ..."
    self.P(header, color='b', boxed=True)

    self._business_manager = BusinessManager(
      log=self.log, 
      owner=self,
      shmem=self.app_shmem,
      run_on_threads=self.cfg_plugins_on_threads,
      environment_variables=self.cfg_plugins_environment,
      DEBUG=self.DEBUG
    )

    self.app_shmem["business_manager"] = self._business_manager
    return

  def _initialize_io_formatter_manager(self):
    header = "Initializing IO formatter manager..."
    self.P(header, color='b', boxed=True)

    self._io_formatter_manager = IOFormatterManager(log=self.log, DEBUG=self.DEBUG)
    formatter_name = self.cfg_io_formatter
    if formatter_name != '':
      # create default (output) formatter
      self._io_formatter_manager.create_formatter(formatter_name)

    self._app_shmem['io_formatter_manager'] = self._io_formatter_manager
    return

  def _initialize_heavy_ops_manager(self):

    header = "Initializing Heavy Operations manager..."
    self.P(header, color='b', boxed=True)

    self._heavy_ops_manager = HeavyOpsManager(
      log=self.log, 
      shmem=self.app_shmem, 
      DEBUG=self.DEBUG,
    )
    self._app_shmem['heavy_ops_manager'] = self._heavy_ops_manager
    return

  def _initialize_file_system_manager(self):
    header = "Initializing File System manager..."
    self.P(header, color='b', boxed=True)

    self._file_system_manager = FileSystemManager(
      log=self.log,
      config=self._config_manager.cfg_app_file_upload,
      DEBUG=self.DEBUG
    )
    self._file_system_manager.validate(raise_if_error=True)
    self._file_system_manager.create_file_system()
    self._app_shmem['file_system_manager'] = self._file_system_manager
    return
  
  
  def _initialize_private_blockchain(self):    
    try:    
      self._blockchain_manager = DefaultBlockEngine(
        log=self.log,
        name=self.e2_id,
        config=self.cfg_blockchain_config,
        eth_enabled=ct.ETH_ENABLED,
      )
    except:
      raise ValueError("Failure in private blockchain setup:\n{}".format(traceback.format_exc()))

    self._app_shmem[ct.BLOCKCHAIN_MANAGER] = self._blockchain_manager
    return
  
  
  def _initialize_r1fs(self):
    try:
      self._r1fs_engine = R1FSEngine(
        logger=self.log,
        debug=self.debug_r1fs,
      )
    except:
      raise ValueError("Failure in r1fs setup:\n{}".format(traceback.format_exc()))
    # end try to initialize R1FS
    self._app_shmem[ct.R1FS_ENGINE] = self._r1fs_engine
    return
  

  @property
  def config_manager(self):
    return self._config_manager

  @property
  def comm_manager(self):
    return self._comm_manager

  @property
  def capture_manager(self):
    return self._capture_manager

  @property
  def serving_manager(self):
    return self._serving_manager

  @property
  def business_manager(self):
    return self._business_manager

  @property
  def file_system_manager(self):
    return self._file_system_manager
  
  @property
  def blockchain_manager(self) -> DefaultBlockEngine:
    return self._blockchain_manager

  @property
  def app_shmem(self):
    return self._app_shmem




from enum import Enum

class AzureLocationEnum(Enum):
  #region Americas
  USWest              = 'westus'
  USWest2             = 'westus2'
  USCentral           = 'centralus'
  USEast              = 'eastus'
  USEast2             = 'eastus2'
  USNorthCentral      = 'northcentralus'
  USSouthCentral      = 'southcentralus'
  USWestCentral       = 'westcentralus'
  CanadaCentral       = 'canadacentral'
  CanadaEast          = 'canadaeast'
  BrazilSouth         = 'brazilsouth'
  #endregion
  
  #region Europe
  EuropeNorth         = 'northeurope'
  EuropeWest          = 'westeurope'
  UKSouth             = 'uksouth'
  UKWest              = 'ukwest'
  FranceCentral       = 'francecentral'
  FranceSouth         = 'francesouth'
  SwitzerlandNorth    = 'switzerlandnorth'
  SwitzerlandWest     = 'switzerlandwest'
  GermanyNorth        = 'germanynorth'
  GermanyWestCentral  = 'germanywestcentral'
  NorwayWest          = 'norwaywest'
  NorwayEast          = 'norwayeast'
  #endregion
  
  #region Asia
  AsiaEast            = 'eastasia'
  AsiaSouthEast       = 'southeastasia'
  JapanEast           = 'japaneast'
  JapanWest           = 'japanwest'
  AustraliaEast       = 'australiaeast'
  AustraliaSouthEast  = 'australiasoutheast'
  AustraliaCentral    = 'australiacentral'
  AustraliaCentral2   = 'australiacentral2'
  IndiaCentral        = 'centralindia'
  IndiaSouth          = 'southindia'
  IndiaWest           = 'westindia'
  KoreaSouth          = 'koreasouth'
  KoreaCentral        = 'koreacentral'
  #endregion
  
  #region Middle East and Africa
  UAECentral          = 'uaecentral'
  UAENorth            = 'uaenorth'
  SouthAfricaNorth    = 'southafricanorth'
  SouthAfricaWest     = 'southafricawest'
  #endregion
  
  #region China
  ChinaNorth          = 'chinanorth'
  ChinaEast           = 'chinaeast'
  ChinaNorth2         = 'chinanorth2'
  ChinaEast2          = 'chinaeast2'
  #endregion
  
  #region German
  GermanyCentral      = 'germanycentral'
  GermanyNorthEast    = 'germanynortheast'
  #endregion


from naeural_core.data.base import DataCaptureThread
from naeural_core.data.mixins_libs import _DataSensorMixin

_CONFIG = {
  **DataCaptureThread.CONFIG,
  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

class StructuredLiveDataCapture(DataCaptureThread,
                                _DataSensorMixin,
                                ):

  CONFIG = _CONFIG


  
  def startup(self):
    # NON-threaded code in startup
    super().startup()
    self._start_sensor()
    self._metadata = self._sensor_metadata
    return
  
  
  def connect(self):
    return self.sensor.maybe_reconnect()
  
    
  def _run_data_aquisition_step(self):
    _obs, _train, _count = self._get_sensor_data()
    
    self._metadata.dataframe_current = _count + 1

    self._add_inputs(
      [
        self._new_input(img=None, struct_data=_obs, metadata=self._metadata.__dict__.copy(), init_data=_train),
      ]
    )
    return 
    
  def _release(self):
    self.sensor.release()
    return  
  
  


import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import io

sns.set(style="ticks", context="talk")
plt.style.use('dark_background')

def plt_to_np(plt, close=True):
  plt.axis('off')
  fig = plt.gcf()
  fig.canvas.draw()
  data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
  w, h = plt.gcf().canvas.get_width_height()
  np_img = data.reshape((int(h), int(w), -1))
  if close:
    plt.close()
  return np_img  


def plt_to_np2(plt, close=True):
  with io.BytesIO() as buff:
      plt.savefig(buff, format='rgba', bbox_inches='tight', pad_inches=0.3)
      buff.seek(0)
      data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
  print('actual size is', np.sqrt(data.shape[0] / 4))
  w, h = plt.gcf().canvas.get_width_height()
  im4c = data.reshape((int(h), int(w), -1))
  im = im4c[:,:,:-1]
  if close:
    plt.close()
  return im  

if __name__ == '__main__':
  
  
  
  x = [1,2,3,4,5]
  y = [4,5,2,4,6]
  
  _ = plt.figure(figsize=(13,8))
  plt.scatter(x,y)
  print(plt.gcf().canvas.get_width_height())
  np_img = plt_to_np(plt)
  plt.imshow(np_img)
  
  



import os
import inspect
import importlib

from ratio1 import BaseDecentrAIObject
class Benchmarker(BaseDecentrAIObject):
  def __init__(self, config, **kwargs):
    self.custom_models_list = []
    
    super().__init__(**kwargs)
    self.read_config(config)
    # self.load_default_models()
    self.load_custom_models()
    self.load_datasets()
    self.load_default_metrics()
    self.load_custom_metrics()
    pass
  
  def read_config(self, config):
    if type(config) == str:
      self.config_dict = self.log.load_json(config)
    elif type(config) == dict:
      self.config_dict = config
    else:
      self.log.P('Invalid config type', c = 'red')
    return
  
  def load_default_models(self):
    for model_path in self.config_dict['MODELS_PATHS']:
      if os.path.isdir(model_path):
        self.load_models_from_dir(model_path)
      elif os.path.isfile(model_path):
        if model_path[-3:] == '.pb':
          self.load_pb_model(model_path)
        if model_path[-3:] == '.h5':
          self.load_h5_model(model_path)
    return
  
  def load_custom_models(self):
    for model_name, model_file in self.config_dict['CUSTOM_MODELS'].items():
      if model_file[-3:] == '.py':
        model_file = model_file[:-3]
      model_file = model_file.replace('/', '.')
      module = importlib.import_module(model_file, package='.')
      classes = inspect.getmembers(module, inspect.isclass)
      _cls = None
      for obj in classes:
        if obj[0] == model_name:
          _cls = obj[1]
          model = _cls(log=self.log)
      self.custom_models_list.append(model)
      
    return
      
  
  def load_datasets(self):
    pass
  
  def load_default_metrics(self):
    pass
  
  def load_custom_metrics(self):
    pass
  
  def benchmark(self):
    pass
  
  def save_results(self):
    pass
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  


from naeural_core.business.training.general_training_process import GeneralTrainingProcessPlugin
from naeural_core.business.training.general_training_process import _CONFIG as BASE_CONFIG

_CONFIG = {
  **BASE_CONFIG,
  'VALIDATION_RULES': {
    **BASE_CONFIG['VALIDATION_RULES'],
  },

  'AUTO_DEPLOY': {},
}


class SecondStageTrainingProcessPlugin(GeneralTrainingProcessPlugin):
  def prepare_inference_plugin_config(self) -> dict:
    return {
      'SECOND_STAGE_MODEL_ARCHITECTURE_PATH': self.training_output['METADATA']['MODEL_ARCHITECTURE_PATH'],
      'SECOND_STAGE_MODEL_HYPERPARAMS': self.training_output['STATUS']['BEST']['model_kwargs'],
      'SECOND_STAGE_CLASS_NAMES': list(self.training_output['METADATA']['CLASSES'].keys()),
      # The following 2 lines
      # 'SECOND_STAGE_MODEL_WEIGHTS_URL': self.cloud_path_to_url(self.weights_uri['CLOUD_PATH']),
      # 'SECOND_STAGE_MODEL_TRACE_URL': self.cloud_path_to_url(self.trace_uri['CLOUD_PATH']),
      'SECOND_STAGE_MODEL_WEIGHTS_URL': self.weights_uri['URL'],
      'SECOND_STAGE_MODEL_TRACE_URL': self.trace_uri['URL'],
      'SECOND_STAGE_INPUT_SIZE': self.training_output['STATUS']['BEST']['dct_grid_option']['input_size'],
      'SECOND_STAGE_TARGET_CLASS': self.training_output['METADATA']['FIRST_STAGE_TARGET_CLASS'],
      'SECOND_STAGE_PREPROCESS_DEFINITIONS': self.training_output['METADATA']['INFERENCE_PREPROCESS_DEFINITIONS'],
      'SECOND_STAGE_OBJECTIVE_NAME': self.cfg_objective_name,
    }

  def auto_deploy(self):
    if self.cfg_auto_deploy is None:
      return
    box_auto_deploy = self.cfg_auto_deploy.get('BOX_ID', None)
    node_address_auto_deploy = self.cfg_auto_deploy.get('NODE_ADDRESS', None)
    streams = self.cfg_auto_deploy.get('STREAMS', [])
    target_classes = self.training_output['METADATA']['FIRST_STAGE_TARGET_CLASS']
    if not isinstance(target_classes, list):
      target_classes = [target_classes]

    for s in streams:
      if 'PLUGINS' not in s:
        s['PLUGINS'] = []

      s['PLUGINS'].append({
        'SIGNATURE': 'second_stage_detection',
        'INSTANCES': [
          {
            'INSTANCE_ID': self.log.now_str(),
            'AI_ENGINE': 'custom_second_stage_detector',
            'OBJECT_TYPE': target_classes,
            'SECOND_STAGE_DETECTOR_CLASSES': list(self.training_output['METADATA']['CLASSES'].keys()),
            'STARTUP_AI_ENGINE_PARAMS': {
              'CUSTOM_DOWNLOADABLE_MODEL_URL': self.inference_config_uri['URL'],  # maybe find way to put
              # cloud path or extend url lifetime (for now the model data is downloaded by the serving manager,
              # which might have a different bucket policy than the training manager)
              'MODEL_INSTANCE_ID': self.training_output['METADATA']['MODEL_NAME'],
            },
            "DESCRIPTION": self.cfg_description,
            "OBJECTIVE_NAME": self.cfg_objective_name,
          }
        ]
      })

      node_address = node_address_auto_deploy or self.netmon.network_node_addr(box_auto_deploy)
      self.cmdapi_start_stream_by_config_on_other_box(node_address=node_address, config_stream=s)
    #endfor

    return



import os
import random

import json
import base64

from copy import deepcopy
from hashlib import sha256
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes

from naeural_core import Logger

if __name__ == '__main__':

  # first the sender process
  # prepare SK, PK, ADDR
  
  # sk = ec.generate_private_key(curve=ec.SECP256K1())
  sk = ec.derive_private_key(private_value=1234, curve=ec.SECP256K1())
  pk = sk.public_key()
  data = pk.public_bytes(
    encoding=serialization.Encoding.X962,
    format=serialization.PublicFormat.CompressedPoint,
  )
  str_address = "0xai_" + base64.urlsafe_b64encode(data).decode()

  # now the actual hasing and signing
  dct_to_send = {'9'  : 9, '2':2, '3':3, '10':{'2':2,'100':100, '1':1}}
  deterministic_json = json.dumps(dct_to_send, sort_keys=True, separators=(',',':'))
  data = deterministic_json.encode()
  str_digest = sha256(data).hexdigest()    
  # bin_digest = str_digest.encode()
  bin_digest = sha256(data).digest()
  
  signature = sk.sign(
      data=bin_digest,
      signature_algorithm=ec.ECDSA(hashes.SHA256()),
  )
  str_signature = base64.urlsafe_b64encode(signature).decode()

  dct_to_send['SIGN'] = str_signature
  dct_to_send['ADDR'] = str_address
  dct_to_send['HASH'] = str_digest

  print("Fully signed object\n{}".format(json.dumps(dct_to_send, indent=2)))  

  if str_address != '0xai_AuN2SENcYNzRgbPUHVCFe6W1q-vieUKap2VY9mU_Fljy':
    Logger.print_color("ERROR: Address generation failed", color='r')
  else:
    Logger.print_color("Deterministic address generation OK", color='g')

  if deterministic_json != '{"10":{"1":1,"100":100,"2":2},"2":2,"3":3,"9":9}':
    Logger.print_color("ERROR: Deterministic JSON generation failed", color='r')
  else:
    Logger.print_color("Deterministic JSON generation OK", color='g')   
  
  if str_digest != '7d72cf5bd6cda16c86dfb2c2c4983464edda5bf78e1ca3a21139f5037454c6f3':
    Logger.print_color("ERROR: Hash generation failed", color='r')
  else:
    Logger.print_color("Hash generation OK", color='g')




import numpy as np

_BASE_ANOMALY_PRC = 0.02

__VER__ = '1.0.1.0'

class BasicAnomalyModel:
  def __init__(self, data_validation_callback=None, **kwargs):
    """
    simple multi-variate Gaussian anomaly detector
    """
    self.trained = False
    self.version = __VER__
    self.data_validation_callback = data_validation_callback # Used to validate if there is sufficient data to train
    return
  
  def _calc_eps_by_prc(self, prc=None, max_eps=0.02):
    """
    compute eps using cut-off percentage
    """
    if prc is None:
      prc = _BASE_ANOMALY_PRC
    pval = self._pdf(self._x_train)
    eps = np.quantile(pval, prc)
    return min(eps, max_eps)
  
  def _calc_simple_proba(self, pvals):
    n = (self._eps - pvals) / self._eps
    p = n * 0.50 + 0.50
    return np.maximum(p, 0)
    
  
  def _calc_eps(self, y_true):
    """
    given a labeled dataset (1 anomaly, 0 not anomaly) we compute the optimal thr 
    without residing on a pre-defined cut-off percentage
    """
    pval = self._pdf(self._train_data) #  we compute pdf value for all observations
    best_eps = 0
    best_F1 = 0
    
    stepsize = (max(pval) -min(pval))/1000
    eps_range = np.arange(pval.min(),pval.max(),stepsize)
    for eps in eps_range: 
      predictions = (pval < eps)[:,np.newaxis] # now we find optimal eps
      tp = np.sum(predictions[y_true==1]==1)   # given that we know which are the outliers
      fp = np.sum(predictions[y_true==0]==1)
      fn = np.sum(predictions[y_true==1]==0)
      
      # compute precision, recall and F1
      prec = tp/(tp+fp)
      rec = tp/(tp+fn)
      
      F1 = (2*prec*rec)/(prec+rec)
      
      if F1 > best_F1:
          best_F1 =F1
          best_eps = eps
      
    return best_eps, best_F1  
  
  @staticmethod
  def calc_pdf(x, loc=None, scale=None):
    if loc is None:
      loc = x.mean()
    if scale is None:
      scale = x.std()
    if len(x.shape) == 1:
      x = x.reshape(-1,1)
    n_feats = x.shape[1]
    if n_feats == 1:
      v = (1 / (scale  * np.sqrt(2 * np.pi))) * np.exp(-0.5 * ((x - loc)/scale)**2)
    else:
      k = n_feats
      sigma2 = np.diag(scale ** 2)
      Xs = x - loc.T
      s_det = np.linalg.det(sigma2)
      s_det_sqrt_inv = 1 / (s_det ** 0.5)
      t1 = 1/((2 * np.pi)**(k/2)) * s_det_sqrt_inv
      t2 = np.exp(-0.5* np.sum(Xs @ np.linalg.pinv(sigma2) * Xs,axis=1))
      v = t1 * t2
    return v  
  
  def _pdf(self, x):   
    """
    Calculates proba density func values given some data and assuming stats are
    already computed
    """
    res = self.calc_pdf(x, loc=self._mean, scale=self._std)
    return res
  
  def fit(self, x_train, y_train=None, prc=None):
    if self.data_validation_callback is not None:
      if not self.data_validation_callback(x_train, y_train):
        # Model will not train; insufficient data
        return
      #endif
    #endif

    if not isinstance(x_train, np.ndarray) and len(x_train.shape) != 2:
      raise ValueError("{} training data must be ndarray with shape [n_obs, n_features] - received {}".format(
        self.__class__.__name__, type(x_train)))
    self._n_feats = x_train.shape[1]
    self._n_obs = x_train.shape[0]
    self._mean = x_train.mean(axis=0)
    self._std = x_train.std(axis=0)  
    self._x_train = x_train
    if y_train is None:
      if prc is None:
        prc = _BASE_ANOMALY_PRC
      _eps = self._calc_eps_by_prc(prc=prc)
    else:
      _eps, _ = self._calc_eps(y_train)
    self._eps = _eps
    self.trained = True
    return
  
  def predict(self, x_test, proba=False):
    if not self.trained: ## TODO: maybe add maybe_predict?
      return []
    if not isinstance(x_test, np.ndarray) or x_test.shape[-1] != self._n_feats:
      raise ValueError("{} input for prediction must be of type np.ndarray [None, N_FEATS]".format(
        self.__class__.__name__))
    x_test = x_test.reshape((-1, self._n_feats))
    pval = self._pdf(x_test)
    if proba:
      return self._calc_simple_proba(pval)
    is_anom = pval <= self._eps
    return is_anom


if __name__ == '__main__':
  from naeural_core.utils.sensors.virtual_sensor import VirtualSensor
  from sklearn.covariance import EllipticEnvelope
  from naeural_core import Logger
  import matplotlib.pyplot as plt
  import seaborn as sns
  sns.set(style="ticks", context="talk")
  plt.style.use('dark_background')

  
  l = Logger('ANOM', base_folder='.', app_folder='_cache', TF_KERAS=False)

  
  for itr in range(1):
    sensor = VirtualSensor(n_max_samples=1000, centers=1)
    x_train = sensor.get_train_data()
    x_test = sensor._data_feed
    n_samples = x_test.shape[0]
    clf = EllipticEnvelope(contamination=0.02)
    clf.fit(x_train)
    
    yh_train1 = clf.predict(x_train) == -1
    yh_test1 = clf.predict(x_test) == -1
  
    model = BasicAnomalyModel()
    model.fit(x_train)
    _yh_train2 = model.predict(x_train, proba=True)
    yh_train2 = _yh_train2 >= 0.5
    _yh_test2 = model.predict(x_test, proba=True)
    yh_test2 = _yh_test2 >= 0.5
    
    
    trn = yh_train1==yh_train2
    tst = yh_test1==yh_test2
    trn_title = "{} Train SKLearn/BasicAnom/match {:>2}/{:>2}/{:>1}".format(
      "GOOD:" if np.all(trn) else "BAD:",
      yh_train1.sum(), yh_train2.sum(), np.all(trn),
      )

    tst_title = "{} Test SKLearn/BasicAnom/match {:>2}/{:>2}/{:>1}".format(
      "GOOD:" if np.all(tst) else "BAD:",
      yh_test1.sum(), yh_test2.sum(), np.all(tst)
      )
    
    
    g_trn = ~yh_train1 & trn     
    a_1 = yh_train1 & (~trn)        
    a_2 = yh_train2 &(~trn)    
    a_12 = yh_train1 & trn
    
    plt.figure(figsize=(13,8))
    plt.scatter(x_train[g_trn][:,0], x_train[g_trn][:,1], marker='o', color='b', label='good')
    plt.scatter(x_train[a_1][:,0], x_train[a_1][:,1], marker='D', color='y', label='clf_bad')
    plt.scatter(x_train[a_2][:,0], x_train[a_2][:,1], marker='x', color='r', label='our_bad')
    plt.scatter(x_train[a_12][:,0], x_train[a_12][:,1], marker='x', color='g', label='both_bad')
    for i, (_x,_y) in enumerate(x_train[yh_train2]):
      plt.annotate("{:.2f}".format(_yh_train2[yh_train2][i]), (_x,_y), fontsize=12)
    for i, (_x,_y) in enumerate(x_train[a_1]):
      plt.annotate("{:.2f}".format(_yh_train2[a_1][i]), (_x,_y), fontsize=12)

    plt.legend()
    plt.title(trn_title)
    plt.show()

    g_tst = ~yh_test1 & tst
    a_1 = yh_test1 & (~tst)        
    a_2 = yh_test2 &(~tst)    
    a_12 = yh_test1 & tst

    plt.figure(figsize=(13,8))
    plt.scatter(x_test[g_tst][:,0], x_test[g_tst][:,1], marker='o', color='b', label='good')
    plt.scatter(x_test[a_1][:,0], x_test[a_1][:,1], marker='D', color='y', label='clf_bad')
    plt.scatter(x_test[a_2][:,0], x_test[a_2][:,1], marker='x', color='r', label='our_bad')
    plt.scatter(x_test[a_12][:,0], x_test[a_12][:,1], marker='x', color='g', label='both_bad')

    for i, (_x,_y) in enumerate(x_test[yh_test2]):
      plt.annotate("{:.2f}".format(_yh_test2[yh_test2][i]), (_x,_y), fontsize=12)
    for i, (_x,_y) in enumerate(x_test[a_1]):
      plt.annotate("{:.2f}".format(_yh_test2[a_1][i]), (_x,_y), fontsize=12)

    plt.title(tst_title)
    plt.legend()
    plt.show()
    
    for i in range(n_samples):
      obs = sensor.get_observation()
      yh1 = clf.predict(obs.reshape((1,-1)))
      yh2 = model.predict(obs)
      if yh1 != yh2:
        yh1 = clf.predict(obs.reshape((1,-1)))
        yh2 = model.predict(obs)
  



#global dependencies
import abc

#local dependencies
from naeural_core import constants as ct

from naeural_core.business.base import CVPluginExecutor
from naeural_core.business.base.cv_plugin_executor import _CONFIG as BASE_CONFIG

_CONFIG = {
  **BASE_CONFIG,
  
  'RUN_WITHOUT_IMAGE' : True,
  'NON_VIDEO_PLUGIN'  : True,  
  'NO_WITNESS'        : False,

  
  
  'VALIDATION_RULES' : {
    **BASE_CONFIG['VALIDATION_RULES'],
  }
}

class BasicStructuredDataPluginExecutor(CVPluginExecutor):
  def __init__(self, **kwargs):

    self.trained = False
    self._counter = 0

    super(BasicStructuredDataPluginExecutor, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    self._init()
    return

  
  def predict(self, data):
    self._counter += 1

    res = self._predict(data=data)
    dct_res, alert_value = ({}, None) if res is None else res
    # is_alert is generated by self.__alert_helper
    self.add_alerter_observation(alert_value)
    return dct_res
  
  def _init(self):
    return

  ### Section A - methods that should be implemented in subclasses
  if True:

    @abc.abstractmethod
    def _train(self, data):
      raise NotImplementedError()

    @abc.abstractmethod
    def _retrain(self, data):
      raise NotImplementedError()

    @abc.abstractmethod
    def _predict(self, data):
      raise NotImplementedError()
  #endif

  ### Section B - methods overwritten from parent
  if True:
    def _process(self):
      train_data = self.dataapi_specific_input_init_data()
      if train_data is not None:
        if not self.trained:
          self.P("Training {}...".format(self.__class__.__name__), color='y')
          self._train(train_data)
          self.P("Done training {}.".format(self.__class__.__name__), color='y')
        else:
          self.P("RE-Training {}...".format(self.__class__.__name__), color='y')
          self._retrain(train_data)
          self.P("Done retraining {}.".format(self.__class__.__name__), color='y')
        #endif
      #endif

      step_data = self.dataapi_struct_data()

      dct_result = self.predict(data=step_data)

      img = None if self._instance_config.get(ct.NO_WITNESS, True) else self.get_witness_image()

      payload = self._create_payload(
        trained=self.trained,
        img=img,
        struct_data_pass_counter=self._counter,
        is_in_alert=self.alerter_is_alert(),
        is_alert=self.alerter_is_new_alert(),
        msg='{} : {}'.format(self.__class__.__name__, self.__version__),
        **dct_result,
      )

      return payload

    def _post_process(self):

      if self._payload is None:
        ### TODO: maybe remove
        if self.cfg_debug_payloads and self.need_refresh():  # if in "demo mode"
          self._payload = self._create_payload(
            img=self.get_witness_image()
          )
          self._add_alert_info_to_payload(self._payload)
        #endif
      #endif

      return
  #endif



"""
{
    "NAME": "UsbCameraMeta",
    "DESCRIPTION": "Test for croppers and other stuff",
    "TYPE": "MetaStream",
    "COLLECTED_STREAMS" : ["UsbCamera"],


    "PLUGINS": [
        {
            "INSTANCES": [
                {
                    "INSTANCE_ID": "WITNES_ON_META",
                    "PROCESS_DELAY": 1
                }
            ],
            "SIGNATURE": "VIEW_SCENE_01"
        }
    ]
}  

"""

from naeural_core.data.base import BaseDataCapture

_CONFIG = {
  **BaseDataCapture.CONFIG,
  'IS_META_STREAM'       : True,
  
  # Parameter that is used to tell the metastream if it should have data from all the collected streams ('IS_LOOSE' = False)
  # or if it can send the aggregated data whenever any collected stream has something ('IS_LOOSE' = True)
  'IS_LOOSE'          : True,
  
  'IS_THREAD'         : False,
  'COLLECTED_STREAMS' : [],
  
  
  'VALIDATION_RULES' : {
    **BaseDataCapture.CONFIG['VALIDATION_RULES'],
    
    'COLLECTED_STREAMS' : {
      'TYPE'    : 'list',
      'MIN_LEN' : 1,
    }
  },
}

class MetaStreamDataCapture(BaseDataCapture):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(MetaStreamDataCapture, self).__init__(**kwargs)
    return
  
  
  def post_process_inputs(self, lst_stream_inputs):
    return lst_stream_inputs
  





import os
import numpy as np
import torch as th

from naeural_core.local_libraries.nn.utils import conv_output_shape, get_model_size
from naeural_core.local_libraries.nn.th.layers import (
  GlobalMaxPool2d,
  InputPlaceholder,
  ReshapeLayer,
  Conv2dExt,
  ConvTranspose2dExt,
  IdentityLayer,
  AffineTransform,
  SpatialTransformerNet,
)


BASIC_ENCODER = [
  {
   "kernel"  : 3,
   "stride"  : 2,
   "filters" : 32,
   "padding" : 1
  },
  {
   "kernel"  : 3,
   "stride"  : 2,
   "filters" : 128,
   "padding" : 1,
  },
  {
   "kernel"  : 3,
   "stride"  : 1,
   "filters" : None, # this will be auto-calculated for last encoding layer
   "padding" : 1,
  },
]

def calc_embed_size(h, w, c, root=3, scale=1):
  img_size = h * w * c
  v = int(np.power(img_size, 1/root))
  v = v * scale
  # now cosmetics
  vf = int(v / 4) * 4
  return vf


class SimpleImageEncoder(th.nn.Module):
  def __init__(self, h, w, channels,
               root=3, scale=1,
               embed_size=None,
               layers=BASIC_ENCODER):
    super().__init__()
    self.hw = (h, w)
    self.layers = th.nn.ModuleList()
    last_channels = channels
    input_shape = (channels, h, w)
    for layer in layers:
      k = layer.get('kernel', 3)
      s = layer.get('stride', 2)
      p = layer.get('padding', 1)
      f = layer['filters']
      if f is None:
        if embed_size is None:
          f = calc_embed_size(
            h, w,
            c=channels,
            root=root,
            scale=scale
            )
        else:
          f = embed_size
      cnv = Conv2dExt(
        input_shape=input_shape,
        in_channels=last_channels,
        out_channels=f,
        kernel_size=k,
        stride=s,
        padding=p,
        show_input=False,
        )
      input_shape = cnv.output_shape
      last_channels = f
      bn = th.nn.BatchNorm2d(f)
      act = th.nn.ReLU()
      self.layers.append(cnv)
      self.layers.append(bn)
      self.layers.append(act)
    self.embed_layer = GlobalMaxPool2d()
    self.encoder_embed_size = last_channels
    return

  def forward(self, inputs):
    th_x = inputs
    for layer in self.layers:
      th_x = layer(th_x)
      # print('  ',th_x.shape)
    th_out = self.embed_layer(th_x)
    return th_out


class SimpleImageDecoder(th.nn.Module):
  def __init__(self,
               h, w,
               channels,
               embed_size=None,
               root=3,
               scale=1,
               layers=BASIC_ENCODER
               ):
    super().__init__()
    if embed_size is None:
      embed_size = calc_embed_size(
        h, w,
        c=channels,
        root=root,
        scale=scale,
        )
    self.hw = (h, w)
    self.layers = th.nn.ModuleList()
    reduce_layers = len([x['stride'] for x in layers if x['stride'] > 1])
    input_layer = InputPlaceholder((embed_size,))
    expansion_channels = int(embed_size ** 0.5)
    expansion_h = h // (2 ** reduce_layers)
    expansion_w = w // (2 ** reduce_layers)
    expansion_size = expansion_h * expansion_w * expansion_channels
    expansion_layer = th.nn.Linear(embed_size, expansion_size)
    reshape_layer = ReshapeLayer((
      expansion_channels,
      expansion_h, expansion_w
      ))
    input_shape = (expansion_channels, expansion_h, expansion_w)
    self.layers.append(input_layer)
    self.layers.append(expansion_layer)
    self.layers.append(reshape_layer)
    last_channels = expansion_channels
    layers = list(reversed(layers))
    for layer in layers:
      k = layer.get('kernel', 3)
      s = layer.get('stride', 2)
      p = layer.get('padding', 1)
      f = layer['filters']
      if f is None:
        f = embed_size
      if s == 1:
        cnv = Conv2dExt(
          input_shape=input_shape,
          in_channels=last_channels,
          out_channels=f,
          kernel_size=k,
          stride=1,
          padding=p,
          show_input=False,
          )
      else:
        cnv = ConvTranspose2dExt(
          in_channels=last_channels,
          input_shape=input_shape,
          out_channels=f,
          kernel_size=k-1,
          stride=s,
          show_input=False,
          # padding=p
          )
      input_shape = cnv.output_shape
      last_channels = f
      bn = th.nn.BatchNorm2d(f)
      act = th.nn.ReLU()
      self.layers.append(cnv)
      self.layers.append(bn)
      self.layers.append(act)

    self.out_layer = Conv2dExt(
      in_channels=last_channels,
      out_channels=channels,
      kernel_size=1,
      input_shape=input_shape,
      show_input=False,
    )
    return

  def forward(self, inputs):
    th_embed = inputs
    th_x = th_embed
    for layer in self.layers:
      th_x = layer(th_x)
      # print('  ',th_x.shape)
    th_out = self.out_layer(th_x)
    return th_out


class SimpleDomainAutoEncoder(th.nn.Module):
  def __init__(self,
               in_h, in_w,
               out_h, out_w, channels,
               domain_name,
               variable_input=False,
               save_folder='_local_cache/models',
               root=3,
               scale=3,
               embed_size=None,
               layers=BASIC_ENCODER
               ):
    super().__init__()

    self.domain_name = domain_name
    self.save_folder = save_folder

    self.encoder = SimpleImageEncoder(
      h=in_h, w=in_w,
      channels=channels,
      layers=layers,
      root=root,
      scale=scale,
      embed_size=embed_size,
      )

    self.decoder = SimpleImageDecoder(
      embed_size=self.encoder.encoder_embed_size,
      h=out_h, w=out_w,
      channels=channels,
      layers=layers
      )
    return

  def forward(self, inputs):
    th_x = self.encoder(inputs)
    th_out = self.decoder(th_x)
    return th_out


  def save_encoder(self, path=None):
    if path is None:
      os.makedirs(self.save_folder, exists_ok=True)
      path = os.path.join(
        self.save_folder,
        "{}_enc{}.pt".format(
          self.domain_name,
          self.encoder.encoder_embed_size
          )
        )
    in_train = self.encoder.training
    self.encoder.eval()
    th.save(self.encoder.state_dict(), path)
    self.encoder_save_path = path
    if in_train:
      self.encoder.train()
    return

  def save_decoder(self, path=None):
    if path is None:
      os.makedirs(self.save_folder, exists_ok=True)
      path = os.path.join(
        self.save_folder,
        "{}_dec{}.pt".format(
          self.domain_name,
          self.encoder.encoder_embed_size
          )
        )
    in_train = self.decoder.training
    self.decoder.eval()
    th.save(self.decoder.state_dict(), path)
    self.decoder_save_path = path
    if in_train:
      self.decoder.train()
    return

class AdvancedSpatialTransformerNet(th.nn.Module):
  def __init__(
      self,
      input_shape:list=[],
      loc_layers=[(8,7), (10,5)],
      drop_input=0,
      drop_fc=0,
      fc_afine=[32],
      reflection=True,
      # below advanced layers
      decode_reshape=None, # TODO: implement (CHW)
      decode_layers=[], # (f,k)
      decode_act=False,
    ) -> None:
    super().__init__()

    assert len(decode_layers) > 0, "Must specify auto-encoder design with `decode_layers"

    self.stn = SpatialTransformerNet(
      input_shape=input_shape,
      loc_layers=loc_layers,
      drop_input=drop_input,
      drop_fc=drop_fc,
      fc_afine=fc_afine,
      reflection=reflection,
      affine_reshape=input_shape,
    )

    channels = input_shape[0]
    if decode_reshape is None:
      decode_reshape = self.stn.affine_reshape

    if len(decode_reshape) == 2:
      decode_reshape = (channels, ) + tuple(decode_reshape)

    decs = []
    prev_output = (input_shape[0],) + self.stn.affine_reshape[1:]
    prev_f = channels
    for f,k in decode_layers:
      decs.append(
        Conv2dExt(
          input_shape=prev_output,
          in_channels=prev_f,
          out_channels=f,
          kernel_size=k,
          stride=2,
      ))
      prev_output = decs[-1].output_shape
      decs.append(th.nn.ReLU(True))
      prev_f = f
    decs.append(th.nn.Upsample(decode_reshape[1:]))
    decs.append(
      Conv2dExt(
        input_shape=decode_reshape[1:],
        in_channels=prev_f,
        out_channels=f,
        kernel_size=1,
      )
    )
    prev_output = decs[-1].output_shape
    decs.append(th.nn.ReLU(True))


    self.decoder = th.nn.ModuleList(decs)
    if decode_act is not None:
      self.readout_image = th.nn.Sequential(
        Conv2dExt(
          input_shape=prev_output,
          in_channels=prev_f,
          out_channels=channels,
          kernel_size=1
        ),
        th.nn.Sigmoid()
      )
    else:
      self.readout_image = Conv2dExt(
        input_shape=prev_output,
        in_channels=prev_f,
        out_channels=channels,
        kernel_size=1
      )
    return


  def forward(self, inputs):
    th_theta, th_x_img = self.stn(inputs)

    for layer in self.decoder:
      th_x_img = layer(th_x_img)

    th_image = self.readout_image(th_x_img)

    return th_theta, th_image


  def transform(self, images, shape):
    th_thetas, _ = self(images)

    th_transform_matrix = th.nn.functional.affine_grid(th_thetas, shape)

    th_transformed = th.nn.functional.grid_sample(images, th_transform_matrix)
    return th_transformed


if __name__ == "__main__":
  TEST_SIMPLE_DOMAIN_AE = False
  TEST_ADV_STN = True

  if TEST_SIMPLE_DOMAIN_AE:
    th_x = th.rand((1,3,120,200))
    model = SimpleDomainAutoEncoder(
      in_h=th_x.shape[2], in_w=th_x.shape[3],
      out_h=120, out_w=200,
      channels=th_x.shape[1],
      embed_size=256,
      domain_name='test'
    )
    print(model)
    th_y = model(th_x)
    print(th_y.shape)

    npars, nbytes = get_model_size(model)
    print('{:,} params, {:,} bytes'.format( npars,nbytes))

  if TEST_ADV_STN:
    mst = AdvancedSpatialTransformerNet(
      input_shape=(1,120,200),
      loc_layers=[(8,7), (10,5)],
      decode_act=True,
      decode_layers=[(16, 5), (32, 3)],
      decode_reshape=(24,94),
    )

    inp = th.rand((1, 1, 120, 200))
    print(mst)
    o = mst(inp)
    print(o[0].shape)
    print(o[1].shape)  


"""
  TODO:
    Add GPU SYNC !
"""


import torchvision.transforms as T

from naeural_core import constants as ct
from naeural_core.serving.default_inference.th_rface import ThRface as ParentServingProcess

_CONFIG = {
  **ParentServingProcess.CONFIG,
  # PARAMETERS USED FOR THE `FACE_ID` MODEL
  'FACEID_MODEL_FILENAME': 'MobileFaceNetV3_large_ArcFace_glint360k_cosface_r100_fp16_0.1_cosine_student.ths',
  'FACEID_MODEL_URL': 'minio:FaceIdentification/MobileFaceNetV3_large_ArcFace_glint360k_cosface_r100_fp16_0.1_cosine_student.ths',
  'FACEID_INPUT_SIZE': (3, 112, 112),
  "MAX_BATCH_SECOND_STAGE": 5,

  "FACEID_IMAGE_HW": (112, 112),
  # USE_AMP"                     : False,

  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

__VER__ = '0.1.0.0'


class ThRfaceId(ParentServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    ver = kwargs.pop('version', None)
    if ver is None:
      ver = __VER__

    super(ThRfaceId, self).__init__(version=ver, **kwargs)
    self._has_second_stage_classifier = True
    # TODO(Adi): choose between resize and resize with pad
    return

  def _startup(self):
    super()._startup()
    self._face_id_transforms = T.Compose([
       T.Resize(size=(self.cfg_faceid_image_hw[0], self.cfg_faceid_image_hw[1])),
        #  PreprocessResizeWithPad(h=self.cfg_faceid_image_hw[0], w=self.cfg_faceid_image_hw[1], normalize=False),
    ])

  def _seconde_stage_aggregate_batch_predict_callback(self, lst_th_x):
    return self.th.cat(lst_th_x) if len(lst_th_x) > 0 else self.th.tensor([])

  def _second_stage_classifier(self, lst_th_faces_cropped):
    """ This will be the stage where we embed faces and return embeddings """

    # get batch size for each image
    result_mapping = [len(x) for x in lst_th_faces_cropped]

    # concat all images to batch predict all images at once
    th_faces_batch = self.th.cat(lst_th_faces_cropped)

    # run a batch predict with all crops
    th_results = self._batch_predict(
      prep_inputs=th_faces_batch,
      model=self.faceid,
      batch_size=self.cfg_max_batch_second_stage,
      aggregate_batch_predict_callback=self._seconde_stage_aggregate_batch_predict_callback
    )

    # extract batch of embeddings for each image
    face_features = []
    for b_size in result_mapping:
      face_features.append(th_results[:b_size])
      th_results = th_results[b_size:]
    return face_features, lst_th_faces_cropped

  def _scale_coords(self, lst_np_pred_nms_cpu):
    lst_np_pred_nms_cpu_denorm = []
    for i in range(len(lst_np_pred_nms_cpu)):
      np_pred_nms_cpu = lst_np_pred_nms_cpu[i].cpu().numpy()
      original_shape = self._lst_original_shapes[i]

      if np_pred_nms_cpu.shape[0] != 0:
        denorm_array = self.np.array(
          [self.cfg_input_size[1], self.cfg_input_size[0]] * 2 +
          [1] +
          [self.cfg_input_size[1], self.cfg_input_size[0]] * 5
        )
        np_pred_nms_cpu_denormed = np_pred_nms_cpu * denorm_array
        np_pred_nms_cpu[:, :4] = self.scale_coords(
          img1_shape=self.cfg_input_size,
          coords=np_pred_nms_cpu_denormed[:, :4],
          img0_shape=original_shape,
        ).round()

        np_pred_nms_cpu[:, 5:] = self.scale_coords(
          img1_shape=self.cfg_input_size,
          coords=np_pred_nms_cpu_denormed[:, 5:],
          img0_shape=original_shape,
        ).round()

        lst_np_pred_nms_cpu_denorm.append(np_pred_nms_cpu)
      else:
        lst_np_pred_nms_cpu_denorm.append([])
    return lst_np_pred_nms_cpu_denorm

  def _crop_faces(self, lst_th_pred_nms, lst_th_image):
    lst_lst_th_face_cropped = []

    # denorm input to be 0-255 uint8
    lst_th_image += self.norm_sub_tensor
    lst_th_image = lst_th_image.to(dtype=self.th.uint8)

    for th_preds, th_image in zip(lst_th_pred_nms, lst_th_image):
      lst_image_crops = []
      for th_pred in th_preds:

        L, T, R, B, P = th_pred[:5].cpu()
        L = int(L * self.cfg_input_size[1])
        R = int(R * self.cfg_input_size[1])
        T = int(T * self.cfg_input_size[0])
        B = int(B * self.cfg_input_size[0])
        H = th_image.shape[1]
        W = th_image.shape[2]

        if P < 0.3:
          # maybe ignore this image
          pass
        # Here we create a new image that is supposed to have the original dimensions of the detected box
        # and fill it with the corresponding data from the original image. If the detection is outside the image,
        # we fill the extra space with black
        th_crop = self.th.zeros((3, B - T, R - L), dtype=th_image.dtype, device=th_image.device)
        th_crop[:, max(0, -T):min(B - T, H - T), max(0, -L): min(R - L, W - L)] = \
            th_image[:, max(0, T): min(B, H), max(0, L): min(R, W)]
        # th_crop = th_image[:, T:B, L:R]
        lst_image_crops.append(th_crop)
      lst_lst_th_face_cropped.append(lst_image_crops)
    return lst_lst_th_face_cropped

  def _second_stage_pre_process(self, pred_nms, th_inputs):
    preds_nms_denormed = self._scale_coords(pred_nms)

    # cut the images according to `preds_nms_denormed`
    lst_lst_th_face_cropped = self._crop_faces(pred_nms, th_inputs)

    """    
    image = cap.read()
    
    image = cv2.resize(image, (96, 96)) 
    img = image[...,::-1]
    img = np.around(np.transpose(img, (2,0,1))/255.0, decimals=12)
    x_train = np.array([img])
    embedding = model.predict(x_train)
    """

    lst_th_faces_cropped = [self.th.stack([self._face_id_transforms(th_face) for th_face in lst_th_face])
                            if len(lst_th_face) > 0 else self.th.tensor([], dtype=self.th.uint8, device=self.dev)
                            for lst_th_face in lst_lst_th_face_cropped
                            ]

    return preds_nms_denormed, lst_th_faces_cropped

  def _retina_post_process(self, th_preds):
    th_pred_nms, th_n_dets = th_preds
    self.log.start_timer('slice_det')
    pred_nms = [x[:th_n_dets[i]] for i, x in enumerate(th_pred_nms)]
    self.log.stop_timer('slice_det')

    return pred_nms

  def _second_stage_process(self, th_preds, th_inputs=None, **kwargs):
    self.log.start_timer('retina_nms')

    pred_nms = self._retina_post_process(th_preds)
    self.log.stop_timer('retina_nms')

    # second stage preprocess
    preds_nms_denormed, lst_th_faces_cropped = self._second_stage_pre_process(pred_nms, th_inputs)

    pred_second = None
    if self.has_second_stage_classifier:

      self.log.start_timer('2nd_stage_clf')
      pred_second = self._second_stage_classifier(lst_th_faces_cropped)
      self.log.stop_timer('2nd_stage_clf')
    # endif

    return preds_nms_denormed, pred_second

  def _post_process(self, preds):
    pred_nms_cpu, (lst_th_embeds, lst_th_crop_faces) = preds
    nr_images = len(self._lst_original_shapes)
    lst_results = []
    for i in range(nr_images):
      # now we have each individual image and we generate all objects
      # what we need to do is to match `second_preds` to image id & then
      # match second clf with each box
      np_pred_nms_cpu = pred_nms_cpu[i]
      np_embed_cpu = lst_th_embeds[i].cpu().numpy()
      np_crop_faces_cpu = lst_th_crop_faces[i].cpu().numpy()

      lst_inf = []
      for det, embed, crop_face in zip(np_pred_nms_cpu, np_embed_cpu, np_crop_faces_cpu):
        det = [float(x) for x in det]
        # order is [left, top, right, bottom, proba] =>
        # [L, T, R, B, P, P1x, P1y, P2x, P2y, P3x, P3y, P4x, P4y, P5x, P5y]
        # where Pix and Piy are coordinates for Pi and
        # Pi is from [right ear, left eat, nose, right mouth corner, left mouth corner]
        L, T, R, B, P = det[:5]  # order is [left, top, right, bottom, proba]
        kpts = det[5:]
        dct_obj = {
          ct.TLBR_POS: [int(T), int(L), int(B), int(R)],
          ct.PROB_PRC: round(float(P), 2),
          ct.TYPE: "FACE",
          "KEYPOINTS": [(int(kpts[i]), int(kpts[i + 1])) for i in range(0, len(kpts), 2)],
          "EMBEDDING": embed,
          "CROP_FACE": crop_face,
        }
        lst_inf.append(dct_obj)
      lst_results.append(lst_inf)
    return lst_results

  def _second_stage_model_load(self):

    self.download(
      url=self.cfg_faceid_model_url,
      fn=self.cfg_faceid_model_filename
    )

    self.faceid = self.th.jit.load(self.log.get_models_file(self.cfg_faceid_model_filename))

    model_dev = next(self.faceid.parameters()).device
    if (  # need complex check as dev1 != dev2 will be true (if one index is None and other is 0 on same CUDA)
        (model_dev.type != self.dev.type) or  # cpu vs cuda
        ((model_dev.type == self.dev.type) and (model_dev.index != self.dev.index))  # cuda but on different onex
    ):
      self.P("Model '{}' loaded & placed on '{}' -  moving model to device:'{}'".format(
        self.__class__.__name__, model_dev, self.dev), color='y')
      self.faceid.to(self.dev)

    self.faceid.eval()
    return

  def _second_stage_model_warmup(self):
    if self.faceid is not None:
      self.model_warmup_helper(
        model=self.faceid,
        input_shape=(3, *self.cfg_faceid_image_hw),
        max_batch_size=self.cfg_max_batch_second_stage,
        model_name='face_id'
      )
    return super()._second_stage_model_warmup()



# simple wrappers
from .nn.tf.layers import OneHotLayer, SqueezeLayer, SliceAxis1Layer, SliceLayer,\
  SinCosEncodingLayer, RepeatElements, SplitLayer

# advanced layers
from .nn.tf import GatedDense, MultiGatedUnit, SmartSeparableConv2D

# losses
from .nn.tf.losses import (
  focal_loss_softmax, focal_loss_sigmoid, K_huber_loss,
  huber_loss, quantile_loss, quantile_loss_05, quantile_loss_95
)

#optimizers
from .nn.tf.optimizers import COCOB

CUSTOM_LAYERS = {
      "OneHotLayer"         : OneHotLayer,
      "SinCosEncodingLayer" : SinCosEncodingLayer,
      "RepeatElements"      : RepeatElements,
      "SplitLayer"          : SplitLayer,
      "SqueezeLayer"        : SqueezeLayer, 
      "GatedDense"          : GatedDense,     
      "SliceAxis1Layer"     : SliceAxis1Layer,
      "SliceLayer"          : SliceLayer,
      "huber_loss"          : huber_loss,
      "K_huber_loss"        : K_huber_loss,
      "quantile_loss"       : quantile_loss,
      "quantile_loss_05"    : quantile_loss_05,
      "quantile_loss_95"    : quantile_loss_95,
      "focal_loss_softmax"  : focal_loss_softmax,
      "focal_loss_sigmoid"  : focal_loss_sigmoid,
      "MultiGatedUnit"      : MultiGatedUnit,
      "SmartSeparableConv2D": SmartSeparableConv2D,

      "COCOB"               : COCOB
    }



from naeural_core.business.base import BasePluginExecutor

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS': True,

  'DEVICE_DEBUG': False,
}


class Device(BasePluginExecutor):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(Device, self).__init__(**kwargs)
    self.__driver_cache = self.defaultdict(lambda: self.NestedDotDict({"timestamp": 0, "data": None}))
    return

  def __is_expired(self, path) -> bool:
    """
    Check if data in cache is expired
    Parameters
    ----------
    path

    Returns
    -------
    bool
    """
    if path in self.__driver_cache:
      if self.__driver_cache[path].timestamp > 0:
        return self.__driver_cache[path].timestamp < self.time()
      return False
    return True

  def __data_prep(self, data):
    """
    Prepare data for dispatch
    Parameters
    ----------
    data : dict

    Returns
    -------
    dict
    """
    if isinstance(data, dict):
      return {k.lower(): v for k, v in data.items()}
    return {}

  def __dispatch_command(self, data, **kwargs):
    """
    Dispatch the command to the right method
    Parameters
    ----------
    data : dict
    kwargs : dict

    Returns
    -------
    any
    """
    dct_actions = self.__data_prep(data)
    action = dct_actions.pop("action", None)
    if isinstance(action, str):
      command = "device_action_{}".format(action.lower())
      func = getattr(self, command, None)
      if func:
        return func(**dct_actions)
      else:
        msg = "Undefined action: {}".format(action)
    else:
      msg = "Invalid action format {}".format(action)
    self._create_error_notification(msg)
    return None

  def _create_error_payload(self, message, command_params=None):
    """
    Create error payload
    Parameters
    ----------
    message

    Returns
    -------

    """
    payload = {"device_error": message, "status": "device_error"}
    self.add_payload_by_fields(
      command_params=command_params, **payload
    )
    return

  def _create_action_payload(self, action, message, **kwargs):
    """
    Create action payload as result of command
    Parameters
    ----------
    action
    message

    Returns
    -------

    """
    payload = {"device_action": action, "device_message": message, "status": "device_action", **kwargs}
    self.add_payload_by_fields(**payload)
    return

  def _create_generic_payload(self, **kwargs):
    """
    Create payload
    Parameters
    ----------
    kwargs : dict

    Returns
    -------
    dict
    """
    payload = {"status": "device_payload", **kwargs}
    self.add_payload_by_fields(**payload)

  def _create_device_state_payload(self, state, **kwargs):
    """
    Create device state payload
    Parameters
    ----------
    kwargs : dict

    Returns
    -------
    dict
    """
    payload = {"status": "device_state", "device_state": state, **kwargs}
    self.add_payload_by_fields(**payload)
    return

  def device_set_cache(self, path, data, seconds):
    """
    Store data in cache
    Parameters
    ----------
    path str
    data any
    seconds int

    Returns
    -------
    NestedDotDict

    """
    self.__driver_cache[path].data = data
    self.__driver_cache[path].timestamp = self.time() + seconds if seconds > 0 else 0
    return self.__driver_cache[path]

  def device_get_cache(self, path):
    """
    Get data from cache
    Parameters
    ----------
    path

    Returns
    -------

    """
    if self.__is_expired(path):
      return None
    return self.__driver_cache[path].data

  def on_command(self, data, **kwargs):
    """
    On command received
    Parameters
    ----------
    data : dict
    kwargs : dict

    Returns
    -------
    any
    """
    return self.__dispatch_command(data, **kwargs)



from time import time, sleep
from naeural_core import constants as ct


class _NotificationsCommMixin(object):

  def __init__(self):
    super(_NotificationsCommMixin, self).__init__()
    return

  def _run_thread_notifications(self):
    self._init()
    bytes_delivered = 1 # force to 1 to trigger the first send
    while True:
      try:
        # check stop
        start_it = time()
        if self._stop:
          # stop command received from outside. stop imediatly
          self.P('`stop` command received. Exiting from `{}._run_thread_notifications`'.format(
            self.__class__.__name__))
          break

        # handle send
        self._maybe_reconnect_send()
        if bytes_delivered > 0:
          to_send = None
          if len(self._send_buff) > 0:
            to_send = self._send_buff.popleft()
            bytes_delivered = 0 if to_send is not None else 1

        if to_send is not None and self.has_send_conn:
          msg_id, msg, ts_added_in_buff = to_send
          msg = self._prepare_message(msg=msg, msg_id=msg_id)
          bytes_delivered = self.send_wrapper(msg)
          self.telemetry_maybe_add_message(msg=msg, ts_added_in_buff=ts_added_in_buff, successful_send=bytes_delivered)
        # endif
        end_it = time()
        loop_time = end_it - start_it
        loop_resolution = self.loop_resolution
        sleep_time = max(1 / loop_resolution - loop_time, 0.00001)
        sleep(sleep_time)
      except Exception as e:
        msg = "Exception in notif comm thread loop. Forcing loop delay extension until the exception does not occur."
        self.P(msg, color='r')
        sleep(5)
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          msg=msg,
          autocomplete_info=True
        )
      # end try-except
    # endwhile

    self._release()
    self.P('`run_thread` finished')
    self._thread_stopped = True
    return



from naeural_core.business.default.web_app.fast_api_web_app import FastApiWebAppPlugin as BasePlugin
from ratio1 import Session

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,

  'SAVE_PERIOD': 60,
  "ALLOW_EMPTY_INPUTS": True,
  
  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },
}


class NaeuralFastApiWebApp(BasePlugin):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self.requests_responses = {}
    self.requests_meta = {}
    self.unsolved_requests = set()
    self.session = None
    self.last_save_time = None
    self.force_persistence = False
    super(NaeuralFastApiWebApp, self).__init__(**kwargs)
    return

  def debug_on_payload(self, sess, node_addr, pipeline, signature, instance, payload):
    if self.cfg_debug_mode and not self.__ignore_signature(signature):
      self.P(f'[Session]Payload received from {node_addr} on signature {signature}: {payload}')
    return

  def on_init(self):
    # !!!This approach, although works, will not be allowed in the future because it's not safe
    # TODO: maybe refactor after the new requests system is done
    self.session = Session(
      name=f'{self.str_unique_identification}',
      config=self.global_shmem['config_communication']['PARAMS'],
      log=self.log,
      bc_engine=self.global_shmem[self.ct.BLOCKCHAIN_MANAGER],
      on_payload=self.debug_on_payload,
    )
    super(NaeuralFastApiWebApp, self).on_init()
    self.__maybe_load_persistence_data()
    return

  def relevant_plugin_signatures(self):
    """
    Returns a list of the relevant plugin signatures for the requests sent by this plugin.

    Returns
    -------
    res : list[str]
        The list of relevant plugin signatures.
        If empty, all the requests will be considered relevant.
    """
    return []

  def __ignore_signature(self, signature):
    if not isinstance(self.relevant_plugin_signatures(), list):
      return False
    if signature is None:
      return True
    if len(self.relevant_plugin_signatures()) > 0:
      relevants_lower = [s.lower() for s in self.relevant_plugin_signatures()]
      return signature.lower() not in relevants_lower
    return False

  def process_response_payload(self, payload_data):
    """
    Processes the payload data received from the network.
    Parameters
    ----------
    payload_data : dict
        The payload data.

    Returns
    -------
    res : dict
        The processed response data.
    """
    return payload_data

  def __process_payload_response(self, payload_data: dict):
    """
    Processes the payload data received from the network and updates the requests_responses dictionary.

    Parameters
    ----------

    payload_data : dict
        The payload data.
    """
    signature = payload_data.get('signature', None)
    # Check if the signature is relevant for this plugin.
    if self.__ignore_signature(signature):
      return
    # Check if the payload is a response to a request sent by this plugin.
    request_id = payload_data.get('request_id', None)
    if request_id is not None and request_id in self.unsolved_requests:
      self.requests_responses[request_id] = self.process_response_payload(self.deepcopy(payload_data))
      self.unsolved_requests.remove(request_id)
    return

  def __preprocess_payload_data(self, payload):
    """
    Preprocesses the payload data received from the network.
    First, it checks if the payload data is encrypted and tries to decrypt it.
    Then, it normalizes the keys to lowercase for easier processing.

    Parameters
    ----------

    payload : dict
        The payload data.

    Returns
    -------

    res : dict
        The preprocessed payload data.
    """
    # In case of encrypted data, try to decrypt it.
    payload_data = self.check_payload_data(payload)
    if payload_data is None:
      return
    # Normalize keys to lowercase for easier processing.
    return {k.lower(): v for k, v in payload_data.items()}

  def handle_received_payload(self, payload_data):
    """
    Handles the received payload data.
    This method should be overridden by the user to define the logic for processing the payload data.

    Parameters
    ----------

    payload_data : dict
        The payload data.
    """
    return

  def __maybe_handle_payloads(self):
    payloads = self.dataapi_struct_datas(as_list=True)
    if self.cfg_debug_mode and len(payloads) > 0:
      self.P(f'Received {len(payloads)} payloads.')
    for payload in payloads:
      preprocessed_payload = self.__preprocess_payload_data(payload)
      if preprocessed_payload is None:
        continue
      self.__process_payload_response(preprocessed_payload)
      self.handle_received_payload(preprocessed_payload)
    # endfor payloads
    return

  def get_request_meta(self, **kwargs):
    return kwargs

  def send_network_request(self, **kwargs):
    return NotImplementedError

  def __register_network_request(self, **kwargs):
    """
    Registers a request and sends it to the network through the send_request method.
    Parameters
    ----------
    kwargs : dict
        The keyword arguments to be passed to the send_request method.

    Returns
    -------
    res : bool
        True if the request was successfully registered, False otherwise.
    msg : str
        The error message in case of failure.
    """
    try:
      if 'request_id' not in kwargs:
        request_id = self.uuid()
        kwargs['request_id'] = request_id
      else:
        request_id = kwargs['request_id']
      request_timeout = kwargs.get('timeout', None)
      self.requests_meta[request_id] = self.get_request_meta(**kwargs)
      self.requests_meta[request_id]['start_time'] = self.time()
      if request_timeout is not None:
        self.requests_meta[request_id]['timeout'] = request_timeout
      self.unsolved_requests.add(request_id)
      self.send_network_request(**kwargs)
      return True, None
    except Exception as e:
      msg = f'Error registering request: {e}'
      self.P(msg, color='r')
      return False, msg

  def register_network_request(self, **kwargs):
    return self.__register_network_request(**kwargs)

  def __maybe_get_network_response(self, request_id):
    """
    Returns the network response for specified request if available.
    Parameters
    ----------
    request_id : str
        The request id.

    Returns
    -------
    res : dict
      The network response
    """
    try:
      if request_id in self.requests_responses:
        return self.requests_responses[request_id]
      # endif request_id in requests_responses
      if request_id in self.unsolved_requests:
        req_meta = self.requests_meta.get(request_id) or {}
        timeout = req_meta.get('timeout', None)
        start_time = req_meta.get('start_time', None)
        if timeout is not None and start_time is not None and self.time() - start_time > timeout:
          self.unsolved_requests.remove(request_id)
          return {'error': f'Request {request_id} timed out! Took longer than {timeout} seconds.'}
        # endif timeout reached
      # endif request_id in unsolved_requests
    except Exception as e:
      self.P(f'Error getting network response: {e}', color='r')
    return None

  def maybe_get_network_response(self, request_id):
    return self.__maybe_get_network_response(request_id)

  """BEGIN PERSISTENCE"""
  if True:
    def webapp_get_persistence_data_object(self):
      """
      Here the user can define a dictionary with the data that needs to be saved in the plugin's cache.
      For example, the user can save the plugin's state as follows:
      return {'state': self.state}
      Returns
      -------
      res : dict - the data object to be saved
      """
      return {}

    def __webapp_persistence_save(self):
      data_obj = self.webapp_get_persistence_data_object()
      if len(data_obj.keys()) > 0:
        self.persistence_serialization_save(obj=data_obj)
      # endif data object not empty
      return

    def maybe_persistence_save(self):
      is_save_time = self.last_save_time is None or (self.time() - self.last_save_time > self.cfg_save_period)
      if self.force_persistence or is_save_time:
        self.last_save_time = self.time()
        self.__webapp_persistence_save()
        self.force_persistence = False
      # endif save time
      return

    def webapp_load_persistence_data_object(self, data):
      """
      Here the user can define the logic to load the data object saved in the plugin's cache.
      For example, the user can update the plugin's state based on the loaded data as follows:
      self.state = data['state']
      Parameters
      ----------
      data : dict - the data object to be loaded
      """
      return

    def __maybe_load_persistence_data(self):
      saved_data = self.persistence_serialization_load()
      if saved_data is not None:
        self.webapp_load_persistence_data_object(saved_data)
      # endif saved data available
      return
  """END PERSISTENCE"""

  def process(self):
    super(NaeuralFastApiWebApp, self).process()
    self.maybe_persistence_save()
    self.__maybe_handle_payloads()
    return



"""Perform test request"""
import pprint

import requests

DETECTION_URL = "http://localhost:5000/v1/object-detection/yolov5s"
TEST_IMAGE = "zidane.jpg"

image_data = open(TEST_IMAGE, "rb").read()

response = requests.post(DETECTION_URL, files={"image": image_data}).json()

pprint.pprint(response)



from naeural_core.business.base import BasePluginExecutor as BaseClass

_CONFIG = {
  **BaseClass.CONFIG,

  'ALLOW_EMPTY_INPUTS': False,
  'TEST_NETMON': False,
  'NETMON_CHECK_INTERVAL': 30,

  'VALIDATION_RULES': {
    **BaseClass.CONFIG['VALIDATION_RULES'],
  },
}


class IotListener01Plugin(BaseClass):
  def on_init(self):
    self.last_netmon_check = None
    return

  def message_prefix(self, struct_data):
    event_type = struct_data.get('EE_EVENT_TYPE', None)
    node_addr = struct_data.get('EE_SENDER', 'MISSING_ADDRESS')
    node_id = struct_data.get('EE_ID', 'MISSING_ID')
    pipeline = struct_data.get('PIPELINE', None)
    signature = struct_data.get('SIGNATURE', None)
    instance_id = struct_data.get('INSTANCE_ID', None)

    return f"{event_type} event from <{node_id}:{node_addr}> on route [{pipeline}, {signature}, {instance_id}]"

  def maybe_test_netmon(self):
    if not self.cfg_test_netmon:
      return
    if self.last_netmon_check is not None and self.time() - self.last_netmon_check < self.cfg_netmon_check_interval:
      return
    self.last_netmon_check = self.time()
    lst_available = self.netmon.available_nodes
    self.P(f"Available nodes: {lst_available}")
    lst_accessible = self.netmon.accessible_nodes
    self.P(f"Accessible nodes: {lst_accessible}")
    known_configs = self.netmon.network_known_configs()
    self.P(f"Known configurations: {self.json_dumps(known_configs, indent=2)}")
    accessible_pipelines = [self.netmon.network_node_pipelines(node) for node in lst_accessible]
    self.P(f"Accessible pipelines: {accessible_pipelines}")
    return

  def process(self):
    self.maybe_test_netmon()
    struct_data = self.dataapi_struct_datas()

    self.P(f"Received {len(struct_data)} events.", boxed=False)

    for idx, event in struct_data.items():
      keys = list(event.keys())
      prefix = self.message_prefix(event)

      msg = f"\t{idx}. {prefix} containing {len(keys)} keys: {keys}."
      # msg += f" Detailed data:\n{struct_data}"
      self.P(msg, boxed=False)
    # endfor events

    return





import abc
import json
import numpy as np
import uuid

from time import time


from collections import deque
from naeural_core import DecentrAIObject
from naeural_core import Logger
from threading import Thread

from naeural_core.bc import DefaultBlockEngine
from naeural_core.main.net_mon import NetworkMonitor
from naeural_core.utils.shm_manager import SharedMemoryManager
from naeural_core.utils.plugins_base.plugin_base_utils import _UtilsBaseMixin
from naeural_core import constants as ct

from naeural_core.comm.mixins import (
  _CommunicationTelemetryMixin,
  _DefaultCommMixin,
  _CommandControlCommMixin,
  _NotificationsCommMixin,
  _HeartbeatsCommMixin,
)

from naeural_core.local_libraries import _ConfigHandlerMixin


RUN_ON_THREAD = True
MAX_MESSAGE_LEN = 1024 * 1024 * 2  # 2MB

_CONFIG = {

  'LOG_SEND_COMMANDS': True,
  
  'ENCRYPTED_COMMS': False,       # if True, all comms are end-to-end encrypted

  'DEBUG_LOG_PAYLOADS': False,
  'DEBUG_LOG_PAYLOADS_PIPELINES': [],
  'DEBUG_LOG_PAYLOADS_SIGNATURES': [],
  'DEBUG_LOG_PAYLOADS_SAVE_FILTER': None,
  'DEBUG_LOG_PAYLOADS_REVALIDATE' : False,

  'DEBUG_COMM_ERRORS': False,

  'VALIDATION_RULES': {

  }
}


class BaseCommThread(
  DecentrAIObject,
  _CommunicationTelemetryMixin,
  _DefaultCommMixin,
  _CommandControlCommMixin,
  _NotificationsCommMixin,
  _HeartbeatsCommMixin,
  _ConfigHandlerMixin,
  _UtilsBaseMixin,
):

  __metaclass__ = abc.ABCMeta
  CONFIG = _CONFIG

  def __init__(self,
               log: Logger, shmem, signature,
               comm_type,
               default_config,
               upstream_config,
               environment_variables=None,
               send_channel_name=None,
               recv_channel_name=None,
               timers_section='comms',
               loop_resolution=50,
               **kwargs
               ):
    self._signature = signature
    self._comm_type = comm_type
    self._config = None
    self._default_config = default_config
    self._upstream_config = upstream_config
    self._timers_section = timers_section
    self._environment_variables = environment_variables or {}
    self.__name__ = Logger.name_abbreviation(self._signature.lower()) + \
        '][' + Logger.name_abbreviation(self._comm_type.lower())

    self.shmem = shmem
    self.shm_manager = SharedMemoryManager(
      dct_shared=shmem,
      stream='DEFAULT',
      plugin=comm_type,
      instance=signature,
      category=ct.SHMEM.COMM,
      linked_instances=None,
      log=log,
    )

    self.has_server_conn = False  # public flag for SERVER connection
    self.has_send_conn = False  # public flag for SEND connection
    self.has_recv_conn = False  # public flag for RECV connection
    self._send_to = None
    self._send_buff = deque(maxlen=ct.COMM_SEND_BUFFER)
    self._send_channel_name = send_channel_name
    self._recv_buff = deque(maxlen=ct.COMM_RECV_BUFFER)
    self._recv_channel_name = recv_channel_name
    
    self.__deque_received_hashes = deque(maxlen=1000)
    
    self._nr_conn_retry_iters = 0
    self._total_conn_fails = 0
    self._error_messages = []
    self._error_times = []
    self._last_activity = None

    self._formatter = None
    self._formatter_name = None
    self._heavy_ops_manager = None
    self._network_monitor: NetworkMonitor = None

    self._msg_id = 0
    self.loop_timings = deque(maxlen=10)

    self._incoming_lens = deque(maxlen=100)
    self._incoming_times = deque(maxlen=100)
    self._outgoing_lens = deque(maxlen=100)
    self._outgoing_times = deque(maxlen=100)
    self._bandwidth_mutex = False

    self._stop = False
    self._thread_stopped = False
    self._thread_logs = deque()
    self._thread = None

    self.loop_resolution = loop_resolution
    super(BaseCommThread, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.P("{} Starting comm '{}' {}".format('=' * 10, self._comm_type, '=' * 10))
    if self._environment_variables is not None and len(self._environment_variables) > 0:
      self.P("Updating config with node environment config....", color='m')
      self._config = self._merge_prepare_config(
        delta_config=self._environment_variables
      )
    self._config = self._merge_prepare_config(
      default_config=self._config
    )

    self.setup_config_and_validate(self._config)

    self._heavy_ops_manager = self.shmem['heavy_ops_manager']
    self._network_monitor: NetworkMonitor = self.shmem['network_monitor']
    self._io_formatter_manager = self.shmem['io_formatter_manager']
    self._formatter, self._formatter_name = self._io_formatter_manager.get_formatter()
    if self._formatter_name is not None:
      self._formatter_name = self._formatter_name.lower()
    return

  @property
  def bc_engine(self) -> DefaultBlockEngine: 
    return self.shmem[ct.BLOCKCHAIN_MANAGER]

  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color, str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.COMM
    super().P(s, prefix=True, color=color, **kwargs)
    return

  @property
  def last_activity_time(self):
    return self._last_activity

  @property
  def server_address(self):
    return self._get_server_address()

  @abc.abstractmethod
  def _init(self):
    # do stuff at thread start
    raise NotImplementedError()

  @abc.abstractmethod
  def _maybe_reconnect_send(self):
    # reconnect SEND connection if required
    raise NotImplementedError()

  @abc.abstractmethod
  def _maybe_reconnect_recv(self):
    # reconnect RECV connection if required
    raise NotImplementedError()

  @abc.abstractmethod
  def _send(self, data, send_to=None):
    # use SEND connection to send data
    raise NotImplementedError()

  @abc.abstractmethod
  def _maybe_fill_recv_buffer(self):
    # get message, decode it and append to self._recv_buffer
    raise NotImplementedError()

  @abc.abstractmethod
  def _release(self):
    # do thread cleanup stuff
    raise NotImplementedError()

  @property
  def cfg_max_retry_iters(self):
    return self._config.get(ct.CONN_MAX_RETRY_ITERS, 3)

  @property
  def comm_failed_after_retries(self):
    if self._nr_conn_retry_iters > self.cfg_max_retry_iters:
      return True
    return False

  def _process_signature(self, signature, payload):
    # add some signature processing if needed.
    # BC related stuff should happen in the BC engine
    return

  def __sign(self, data):
    if self.bc_engine is not None:
      self.start_timer('bc_sign')
      # if inplace=True, it will modify the original data and SAVE deepcopy time
      prepared_data = self.log.replace_nan(data, inplace=True)
      # TODO: should serialize sets
      # prepared_data = self.log.serialize_sets(prepared_data)
      signature = self.bc_engine.sign(prepared_data, add_data=True, use_digest=True)
      self._process_signature(signature=signature, payload=prepared_data)
      self.end_timer('bc_sign')      
      # self.P("DEBUG:\n{}".format(self.log.safe_dumps_json(data, indent=4)))            
    else:
      prepared_data = data
    return prepared_data

  def add_incoming(self, size):
    if self._bandwidth_mutex:
      return
    self._incoming_lens.append(size)
    self._incoming_times.append(time())
    return

  def add_outgoing(self, size):
    if self._bandwidth_mutex:
      return
    self._outgoing_lens.append(size)
    self._outgoing_times.append(time())
    return

  def get_incoming_bandwidth(self):
    result = 0
    self._bandwidth_mutex = True
    if len(self._incoming_lens) > 2:
      interval = self._incoming_times[-1] - self._incoming_times[0]
      result = round(np.sum(self._incoming_lens) / interval / 1024, 2)
    self._bandwidth_mutex = False
    return result

  def get_outgoing_bandwidth(self):
    result = 0
    self._bandwidth_mutex = True
    if len(self._outgoing_lens) > 2:
      interval = self._outgoing_times[-1] - self._outgoing_times[0]
      result = round(np.sum(self._outgoing_lens) / interval / 1024, 2)
    self._bandwidth_mutex = False
    return result

  def _get_server_address(self):
    return

  def _get_errors(self):
    return [], []

  def get_error_report(self):
    errors, times = self._get_errors()
    errors = errors + self._error_messages
    times = times + self._error_times
    return errors, times

  def payload_debugger(self, jsonpayload):
    """This method works in conjunction with the `DEBUG_LOG_PAYLOADS` config flag as follows:
      - if `DEBUG_LOG_PAYLOADS` is False, this method is not called
      - if `DEBUG_LOG_PAYLOADS_PIPELINES` has values, we only show payloads from those pipelines
      - if `DEBUG_LOG_PAYLOADS_SIGNATURES` has values, we only show payloads from those signatures


    Parameters
    ----------
    jsonpayload : dict
        the payload that leaves the execution engine.
    """
    pipelines = [x.upper() for x in self.cfg_debug_log_payloads_pipelines]  # constraint pipelines to upper
    signatures = [x.upper() for x in self.cfg_debug_log_payloads_signatures]  # constrain signatures to upper
    
    # payload filter save section
    str_filter = self.cfg_debug_log_payloads_save_filter            
    data = json.loads(jsonpayload)
    if isinstance(str_filter, str) and str_filter in jsonpayload:
      self._save_raw_payload(jsonpayload, prefix='')
    #endif save filter
    # end payload filter save section
    
    show = True
    try:    
      has_action = data.get(ct.COMMS.COMM_SEND_MESSAGE.K_ACTION, None) is not None
      if has_action:
        # this is a SEND command message
        pass
      else:
        pipe = data[ct.PAYLOAD_DATA.EE_PAYLOAD_PATH][1]
        sign = data[ct.PAYLOAD_DATA.EE_PAYLOAD_PATH][2]
        pipe = pipe or ""
        sign = sign or ""
        if len(pipelines) > 0:  # if we have constraint pipelines, we only show those
          show = pipe.upper() in pipelines
        if len(signatures) > 0:  # if we have constraint signatures, we only show those
          show = sign.upper() in signatures
        if show:
          self.P("{}: {}".format(
            len(jsonpayload),
            data[ct.PAYLOAD_DATA.EE_PAYLOAD_PATH])
          )
        #endif show
      #endif send command vs payload
    except:
      self.P("Failed debug: {}".format(data), color='error')
    return
  
  def _check_send_message(self, message):
    """
    This method is called before sending the message to the network.

    Parameters
    ----------
    message : str
        message to be validated.
    """
    
    if len(message) > MAX_MESSAGE_LEN:
      return False
    return True
    
    

  def send_wrapper(self, data, send_to: str = None):
    """
    This is the "final exit" function that gets called when the only thing left is to
    json-ify the message and send it to the network with the custom `_send`

    Parameters
    ----------
    data : dict (usually) although could be a list
      the data dictionary.
    send_to : str
      The recipient of the message. This is used in case we want to send the message on a queue that
      is not listened to by all the participants.
      This will only have effect on the formatable communication topics (e.g. root/{}/config).
      For fixed topics, this will be ignored.

    Returns
    -------
    is_ok : int
      below 0 if error or the len of the message otherwise.

    """
    is_ok = 0
    try:
      # next step will add the signature, hash, addr and also cleanup the payload
      signed_data = self.__sign(data)
      # transform to json
      message = self._jsonify(signed_data)
      
      # now use custom send
      if self.cfg_debug_log_payloads_revalidate:
        # debug area for revalidation
        msg_id = signed_data.get(ct.PAYLOAD_DATA.EE_MESSAGE_ID)
        if msg_id is None:
          str_init = signed_data.get(ct.PAYLOAD_DATA.INITIATOR_ID, "UnkSender")
          str_id = signed_data[ct.PAYLOAD_DATA.EE_ID]
          str_action = signed_data.get("ACTION", "Unknown")
          msg_id = f"{str_init}:{str_action}:{str_id}"
        result = self.bc_engine.verify(signed_data)
        if not result.valid:
          self.P("Failed revalidation on dict for {}: {}".format(msg_id, result.message), color='error')
          self._save_raw_payload(signed_data, prefix='_err-d-', pickle=True)
        #endif valid 
        json_data = json.loads(message)
        result = self.bc_engine.verify(json_data)
        if not result.valid:
          self.P("Failed revalidation on json for {}: {}".format(msg_id, result.message), color='error')
          self._save_raw_payload(message, prefix='_err-j-')
        #endif not valid         
      #endif debug area for revalidate
      
      # now check the message size
      if self._check_send_message(message): 
        self._send(message, send_to=send_to)
        is_ok = len(message)
        if is_ok > 0 and self.cfg_debug_log_payloads:
          self.payload_debugger(message)
        self.add_outgoing(is_ok)
        self._last_activity = time()
      else:
        self.P("Message size {:,.1f} KB was dropped for pipeline {}".format(
          len(message) / 1024, signed_data.get(ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, "<Unknown path>")
          ), color='r'
        )
        is_ok = 1 # we return 1 to signal that the message was dropped and should not be preserved
    except Exception as exc:
      self.has_send_conn = False
      msg = "`send_wrapper` error on payload {} -> {}".format(
        data.get(ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, "<Unknown path>"), exc
      )
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        displayed=True,
        autocomplete_info=True
      )
    return is_ok

  def maybe_fill_recv_buffer_wrapper(self):
    try:
      self._maybe_fill_recv_buffer()
    except Exception as e:
      self.has_recv_conn = False
      msg = "maybe_fill_recv_buffer_wrapper err: {}".format(e)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        autocomplete_info=True
      )
    return

  def _jsonify(self, data):
    """
    From data to json. 
    """
    # no need to `replace_nan=True` as the payload is supposed to be already cleaned up
    message = self.log.safe_dumps_json(data, replace_nan=False, ensure_ascii=False)
    return message

  def start(self):
    run_thread_func = None
    if self._comm_type in [ct.COMMS.COMMUNICATION_DEFAULT, "L_" + ct.COMMS.COMMUNICATION_DEFAULT]:
      run_thread_func = self._run_thread_default
    elif self._comm_type in [ct.COMMS.COMMUNICATION_COMMAND_AND_CONTROL, "L_" + ct.COMMS.COMMUNICATION_COMMAND_AND_CONTROL]:
      run_thread_func = self._run_thread_command_and_control
    elif self._comm_type in [ct.COMMS.COMMUNICATION_HEARTBEATS, "L_" + ct.COMMS.COMMUNICATION_HEARTBEATS]:
      run_thread_func = self._run_thread_heartbeats
    elif self._comm_type in [ct.COMMS.COMMUNICATION_NOTIFICATIONS, "L_" + ct.COMMS.COMMUNICATION_NOTIFICATIONS]:
      run_thread_func = self._run_thread_notifications

    if RUN_ON_THREAD:
      self._thread = Thread(
        target=run_thread_func,
        args=(),
        name=ct.THREADS_PREFIX + 'comm' + run_thread_func.__name__.replace('_run_thread', ''),
        daemon=True,
      )
      self._thread.daemon = True
      self._thread.start()
    else:
      self.P("Warning! Communication Thread does not run on thread", color='e')
      run_thread_func()
    return

  def stop(self):
    self._stop = True
    self._thread.join()
    self._send_buff.clear()
    self._recv_buff.clear()
    self.P('Received `stop` command. Cleaned everything.')
    return

  def get_logs(self):
    logs = []
    while len(self._thread_logs) > 0:
      msg, color = self._thread_logs.popleft()
      logs.append((msg, color))
    return logs
  
  def already_received(self, data):
    payload_hash = data.get(ct.PAYLOAD_DATA.EE_HASH, None)
    initiator_id = data.get(ct.PAYLOAD_DATA.INITIATOR_ID, None)
    initiator_addr = data.get(ct.PAYLOAD_DATA.EE_SENDER, None)
    action = data.get(ct.COMMS.COMM_SEND_MESSAGE.K_ACTION, None)
    already_received = False    
    if payload_hash is not None:
      already_received = payload_hash in self.__deque_received_hashes
      if not already_received:
        self.__deque_received_hashes.append(payload_hash)
      else:
        self.P("Received from {}:{} duplicate '{}' payload with hash: {}".format(
          initiator_id, initiator_addr, action, payload_hash), color='error'
        )
    return already_received
    

  def get_message(self):
    json_msg = None
    if len(self._recv_buff) > 0:
      str_msg = self._recv_buff.popleft()
      incoming_len = len(str_msg)
      self.add_incoming(incoming_len)
      try:
        json_msg = json.loads(str_msg)
        self._last_activity = time()
      except:
        self.P("Cannot decode received message '{}'".format(str_msg), color='r')
        json_msg = None
      #endif json_msg
    #endif len
    if json_msg is not None and self.already_received(json_msg):
      json_msg = None
    return json_msg

  def get_messages(self):
    l = []
    while True:
      msg = self.get_message()
      if msg is None:
        break
      l.append(msg)
    return l

  def get_message_count(self):
    return self._msg_id

  def send(self, data):
    self._msg_id += 1
    msg_id = self._msg_id
    now = self.log.now_str(nice_print=True, short=False)
    self._send_buff.append((msg_id, data, now))
    return

  def _save_raw_payload(self, msg, prefix='', pickle=False):
    try:
      if pickle:
        data = msg
      else:
        data = json.loads(msg)
      msg_id = data[ct.PAYLOAD_DATA.EE_MESSAGE_ID]
      folder = self.log.get_data_subfolder('payloads', force_create=True)
      ext = 'pkl' if pickle else 'txt'
      fn = self.os_path.join(folder, f'{prefix}{msg_id}.{ext}')
      if pickle:
        import pickle
        with open(fn, 'wb') as fhandle:
          pickle.dump(msg, fhandle, protocol=pickle.HIGHEST_PROTOCOL)
      else:
        with open(fn, 'w') as f:
          f.write(msg)
      self.P(f"  Saved payload {msg_id} to {fn}")
    except Exception as exc:
      self.P("Failed raw payload save {}: {}".format(exc, msg), color='r')
    return

  def _save_formatted_payload(self, event_type, signature, formatted_msg):
    fn = '{}_{}_{}.txt'.format(self.log.now_str(), event_type, signature)
    self.log.save_output_json(
      data_json=formatted_msg,
      fname=fn,
      subfolder_path='payloads',
      verbose=False
    )
    return

  def _prepare_message(self, msg, msg_id):
    """
    Final dictionary preparation. This is called in `_run_thread_xxx` together with `send_wrapper`

    Parameters
    ----------
    msg : dict
      the upstream payload dict.
    msg_id : TBD
      TBD.

    Returns
    -------
    msg : dict
      final 'ready-to-json-ify' dict.
      
      
    TODO: review & optimize this method

    """
    WARNING_TIME = 0.1
    dct_original = self.deepcopy(msg) # just to cleanup any possible references still forgotten in the payload
    
    ee_encrypted_payload = dct_original.get(
      ct.PAYLOAD_DATA.EE_IS_ENCRYPTED.lower(), 
      dct_original.get(ct.PAYLOAD_DATA.EE_IS_ENCRYPTED.upper())
    )
    
    ee_encrypted_payload = bool(ee_encrypted_payload) or self.cfg_encrypted_comms
    destination_addr = dct_original.get(ct.PAYLOAD_DATA.EE_DESTINATION)
    destination_id = dct_original.get(ct.PAYLOAD_DATA.EE_DESTINATION_ID) # only for logging atm
    
    ee_id = dct_original.get(ct.EE_ID, None)
    session_id = dct_original.get(ct.PAYLOAD_DATA.SESSION_ID, None)
    
    initiator_id = dct_original.get(ct.PAYLOAD_DATA.INITIATOR_ID, None)
    initiator_addr = dct_original.get(ct.PAYLOAD_DATA.INITIATOR_ADDR, None)
    
    modified_by_id = dct_original.get(ct.PAYLOAD_DATA.MODIFIED_BY_ID, None)
    modified_by_addr = dct_original.get(ct.PAYLOAD_DATA.MODIFIED_BY_ADDR, None)
    
    stream_name = dct_original.get(ct.PAYLOAD_DATA.STREAM_NAME, None)
    signature = dct_original.get(ct.PAYLOAD_DATA.SIGNATURE, None)
    instance_id = dct_original.get(ct.PAYLOAD_DATA.INSTANCE_ID, None)
    payload_path = [ee_id, stream_name, signature, instance_id]
    sequence_number = msg_id  # sequence number for each communicator
    
    # TODO: delete below
    message_id = str(uuid.uuid4())  # unieuq payload ID
    
    ee_timestamp = self.log.now_str(nice_print=True, short=False)
    ee_timezone = self.log.utc_offset
    ee_tz = self.log.timezone
    ee_version = dct_original.get(ct.PAYLOAD_DATA.EE_VERSION, None)
    ee_event_type = dct_original.get(ct.PAYLOAD_DATA.EE_EVENT_TYPE, None)

    # now lets re-create the payload
    dct_outgoing = {
      ct.PAYLOAD_DATA.EE_TIMESTAMP: ee_timestamp,
      ct.PAYLOAD_DATA.EE_TIMEZONE: ee_timezone,
      ct.PAYLOAD_DATA.EE_TZ: ee_tz,
      ct.PAYLOAD_DATA.EE_MESSAGE_SEQ: sequence_number, # TODO: delete
      ct.PAYLOAD_DATA.EE_MESSAGE_ID: message_id,  # TODO: delete (left here for backward compatibility)
      ct.PAYLOAD_DATA.EE_TOTAL_MESSAGES: self.get_message_count(),
      **dct_original
    }

    formatter_name = None
    formatter = None

    if ct.PAYLOAD_DATA.EE_FORMATTER in dct_outgoing:
      # the user defined a custom formatter
      formatter_name = dct_outgoing[ct.PAYLOAD_DATA.EE_FORMATTER]
      formatter = self._io_formatter_manager.get_formatter_by_name(formatter_name)

    elif self._formatter is not None:
      formatter_name = self._formatter_name
      formatter = self._formatter

    if formatter is not None:
      max_elapsed = WARNING_TIME
      dct_outgoing, elapsed = formatter.encode_output(dct_outgoing)
      if elapsed >= max_elapsed:
        self.P(
          "Warning! Formatter time above {} for {}: {:.3f}s".format(
            max_elapsed,
            payload_path, elapsed),
          color='r'
        )
      # endif log too-much-time in formatter
    # endif we have formatter

    dct_outgoing[ct.PAYLOAD_DATA.EE_FORMATTER] = formatter_name    
    dct_output = {}
    
    if destination_addr is None:
      destination_addr, destination_id = None, None
      if initiator_addr is not None:
        destination_addr = initiator_addr
        destination_id = initiator_id
      else:
        destination_addr = modified_by_addr
        destination_id = modified_by_id
    #endif destination    
    if ee_encrypted_payload and destination_addr is not None:
      # encrypt the payload
      str_data = self.log.safe_json_dumps(dct_outgoing, ensure_ascii=False)
      str_enc_data = self.bc_engine.encrypt(
        plaintext=str_data,
        receiver_address=destination_addr, # this can be a address or a list of addresses
      )
      dct_output[ct.PAYLOAD_DATA.EE_ENCRYPTED_DATA] = str_enc_data
      dct_output[ct.PAYLOAD_DATA.EE_IS_ENCRYPTED] = True
      dct_output[ct.PAYLOAD_DATA.EE_DESTINATION] = destination_addr
      if self.cfg_debug_log_payloads:
        self.P("Encrypted {} for '{}' <{}>".format(
          payload_path, destination_id, destination_addr))
    else:
      # just copy data
      dct_output = {
        **dct_outgoing,
        **dct_output,
      }
      dct_output[ct.PAYLOAD_DATA.EE_IS_ENCRYPTED] = False
      if ee_encrypted_payload:
        dct_output[ct.PAYLOAD_DATA.EE_ENCRYPTED_DATA] = "ERROR: No receiver address found!"
    # endif encrypted payload

           
    dct_output[ct.PAYLOAD_DATA.EE_PAYLOAD_PATH] = payload_path    
    
    dct_output[ct.PAYLOAD_DATA.EE_VERSION] = ee_version
    dct_output[ct.PAYLOAD_DATA.EE_EVENT_TYPE] = ee_event_type
    dct_output[ct.PAYLOAD_DATA.SESSION_ID] = session_id
    
    dct_output[ct.PAYLOAD_DATA.INITIATOR_ID] = initiator_id
    dct_output[ct.PAYLOAD_DATA.INITIATOR_ADDR] = initiator_addr
    
    dct_output[ct.PAYLOAD_DATA.MODIFIED_BY_ID] = modified_by_id
    dct_output[ct.PAYLOAD_DATA.MODIFIED_BY_ADDR] = modified_by_addr
    
    dct_output[ct.PAYLOAD_DATA.EE_TIMESTAMP] = ee_timestamp
    dct_output[ct.PAYLOAD_DATA.EE_TIMEZONE] = ee_timezone
    dct_output[ct.PAYLOAD_DATA.EE_TZ] = ee_tz

    ## TODO: DELETE BELOW !
    ### START BACKWARD COMPATIBLE ###
    dct_output[ct.PAYLOAD_DATA.SB_IMPLEMENTATION] = formatter_name
    dct_output[ct.PAYLOAD_DATA.EE_MESSAGE_ID] = message_id # this must be changed to hash after BC is fully operational
    ### END BACKWARD COMPATIBLE ###
    
    # next the EE_HASH, EE_SIGN and EE_SENDER will be added by the BC engine
    
    return dct_output

  def _prepare_command(self, command, receiver_address=None):
    try:
      critical_data = {
        ct.COMMS.COMM_SEND_MESSAGE.K_ACTION: command.pop(ct.COMMS.COMM_SEND_MESSAGE.K_ACTION),
        ct.COMMS.COMM_SEND_MESSAGE.K_PAYLOAD: command.pop(ct.COMMS.COMM_SEND_MESSAGE.K_PAYLOAD),
      }
    except Exception as exc:
      self.P("Failed to prepare command {}:\n{}".format(exc, self.json_dumps(command, indent=3)), color='r')
      return None

    encrypt_payload = self.cfg_encrypted_comms

    if encrypt_payload and receiver_address is not None:
      # encrypt the payload
      str_data = self.log.safe_json_dumps(critical_data, ensure_ascii=False)
      str_enc_data = self.bc_engine.encrypt(
        plaintext=str_data,
        receiver_address=receiver_address,
        #compressed=True, # default is True
        #embed_compressed=True, # default is True - data will be 13:end instead of 12:end due to compression flag
      )
      critical_data = {
        ct.COMMS.COMM_SEND_MESSAGE.K_EE_IS_ENCRYPTED: True,
        ct.COMMS.COMM_SEND_MESSAGE.K_EE_ENCRYPTED_DATA: str_enc_data,
      }
    else:
      critical_data[ct.COMMS.COMM_SEND_MESSAGE.K_EE_IS_ENCRYPTED] = False
      if encrypt_payload:
        critical_data[ct.COMMS.COMM_SEND_MESSAGE.K_EE_ENCRYPTED_DATA] = "ERROR: No receiver address found!"
    # endif encrypted payload

    dct_outgoing = {
      **command,

      ct.COMMS.COMM_SEND_MESSAGE.K_SENDER_ADDR: self.bc_engine.address,

      **critical_data,
    }

    return dct_outgoing



import os
import numpy as np

from datetime import datetime as dt

class _MatplotlibMixin(object):
  """
  Mixin for matplotlib functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_MatplotlibMixin, self).__init__()
    return

  @staticmethod
  def add_copyright_to_plot(plt, push_right=False, annotation=None):
    if annotation is None:
      annotation = "Product Dynamics"

    if push_right:
      _right = 1.2
    else:
      _right = 1.05

    plt.annotate(
      annotation,
      xy=(_right, -0.07), xytext=(0, 2),
      xycoords=('axes fraction'),
      textcoords='offset points',
      size=9, ha='right', va='top',
      bbox=dict(boxstyle="round,pad=0.3", fc="lightgray", ec="lightgray", lw=1),
    )

    return

  def save_plot(self, plt, label='',
                include_prefix=True,
                just_label=False,
                full_path=None):
    """
    saves current figure to file
    """
    _, short_file = self.output_pyplot_image(
      plt=plt,
      label=label,
      include_prefix=include_prefix,
      just_label=just_label,
      full_path=full_path
    )

    return short_file

  def output_pyplot_image(self, plt, label='',
                          include_prefix=True,
                          use_single_prefix=True,
                          just_label=False,
                          full_path=None):
    """
    saves current figure to a file
    """
    if full_path is None:
      if include_prefix:
        file_prefix = dt.now().strftime("%Y%m%d_%H%M%S")
        if use_single_prefix:
          file_prefix = self.file_prefix
        part_file_name = "FIG_{}_{}{}".format(file_prefix, label, ".png")
      else:
        file_prefix = ""
        part_file_name = "{}{}".format(label, ".png")

      file_name = os.path.join(self._outp_dir, part_file_name)
    else:
      file_name = full_path

    _folder, _fn = os.path.split(file_name)
    self.verbose_log("Saving pic '{}' in [..{}]".format(_fn, _folder[-30:]))
    plt.savefig(file_name)
    return file_name, _fn

  def output_image(self, arr, label=''):
    """
    saves array to a file as image
    """
    label = label.replace(">", "_")
    file_prefix = dt.now().strftime("%Y%m%d_%H%M%S_")
    if ".png" not in label:
      label += '.png'
    file_name = os.path.join(self._outp_dir, file_prefix + label)
    self.verbose_log("Saving {} to figure [...{}]".format(
      arr.shape, file_name[-40:]))
    if os.path.isfile(file_name):
      self.verbose_log("Aborting image saving. File already exists.")
    else:
      import matplotlib.pyplot as plt
      plt.imsave(file_name, arr)
    return file_name

  def save_image(self, **kwargs):
    return self.output_image(**kwargs)

  @staticmethod
  def grid_plot_images(images, labels, is_matrix=False, fig_dim=None, title=None):
    import matplotlib.pyplot as plt
    n_images = len(images)
    if fig_dim is None:
      rows = int(np.round(np.sqrt(n_images)))
      columns = int(rows + 1 * (n_images != rows ** 2))
    else:
      rows, columns = fig_dim
    fig = plt.figure(figsize=(columns * 3, rows * 3))
    axs = []
    for i in range(len(images)):
      ax = fig.add_subplot(rows, columns, i + 1)
      axs.append(ax)
      if is_matrix:
        ax.matshow(images[i])
      else:
        ax.imshow(images[i])
      ax.set_title(labels[i])
    if title is not None:
      fig.suptitle(title, fontsize=30)
    plt.subplots_adjust(wspace=0.1, hspace=0.1)
    plt.show()
